@@ -1,32 +1,32 @@
 """Convert ASTs into L{pydoctor.model.Documentable} instances."""
 
-from __future__ import print_function
-
 import ast
 import sys
+from functools import partial
+from inspect import Parameter, Signature
 from itertools import chain
+from pathlib import Path
+from typing import (
+    Any, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Tuple, Union
+)
 
 import astor
 from pydoctor import epydoc2stan, model
-from six import string_types
 
 
-def parseFile(path):
+def parseFile(path: Path) -> ast.AST:
     """Parse the contents of a Python source file."""
     with open(path, 'rb') as f:
         src = f.read() + b'\n'
-    return parse(src)
+    return _parse(src)
 
-_parse_kwargs = {}
 if sys.version_info >= (3,8):
-    _parse_kwargs['type_comments'] = True
-
-def parse(buf):
-    """Parse the contents of a Unicode or bytes string."""
-    return ast.parse(buf, **_parse_kwargs)
+    _parse = partial(ast.parse, type_comments=True)
+else:
+    _parse = ast.parse
 
 
-def node2dottedname(node):
+def node2dottedname(node: Optional[ast.expr]) -> Optional[List[str]]:
     parts = []
     while isinstance(node, ast.Attribute):
         parts.append(node.attr)
@@ -43,7 +43,7 @@ def node2dottedname(node):
     return parts
 
 
-def node2fullname(expr, ctx):
+def node2fullname(expr: Optional[ast.expr], ctx: model.Documentable) -> Optional[str]:
     dottedname = node2dottedname(expr)
     if dottedname is None:
         return None
@@ -55,7 +55,7 @@ def node2fullname(expr, ctx):
 
 
 class ModuleVistor(ast.NodeVisitor):
-    def __init__(self, builder, module):
+    def __init__(self, builder: 'ASTBuilder', module: model.Module):
         self.builder = builder
         self.system = builder.system
         self.module = module
@@ -144,7 +144,7 @@ class ModuleVistor(ast.NodeVisitor):
         self.default(node)
         self.builder.popClass()
 
-    def visit_ImportFrom(self, node):
+    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
         if not isinstance(self.builder.current, model.CanContainImportsDocumentable):
             self.builder.warning("processing import statement in odd context",
                                  str(self.builder.current))
@@ -156,65 +156,100 @@ class ModuleVistor(ast.NodeVisitor):
 
         if node.level:
             # Relative import.
-            parentMod = self.builder.current.parentMod
+            parent: Optional[model.Documentable] = self.builder.current.parentMod
             for _ in range(node.level):
-                if parentMod is None:
+                if parent is None:
                     self.builder.warning("relative import level too high",
                                          str(node.level))
                     return
-                parentMod = parentMod.parent
-            if parentMod is not None:
-                modname = parentMod.fullName() + '.' + modname
+                parent = parent.parent
+            if parent is not None:
+                modname = f'{parent.fullName()}.{modname}'
+
+        if node.names[0].name == '*':
+            self._importAll(modname)
+        else:
+            self._importNames(modname, node.names)
+
+    def _importAll(self, modname: str) -> None:
+        """Handle a C{from <modname> import *} statement."""
 
         mod = self.system.getProcessedModule(modname)
-        if mod is not None:
-            assert mod.state in [model.ProcessingState.PROCESSING,
-                                 model.ProcessingState.PROCESSED]
-            expandName = mod.expandName
-        else:
-            expandName = lambda name: modname + '.' + name
+        if mod is None:
+            # We don't have any information about the module, so we don't know
+            # what names to import.
+            self.builder.warning("import * from unknown", modname)
+            return
+
+        self.builder.warning("import *", modname)
+
+        # Get names to import: use __all__ if available, otherwise take all
+        # names that are not private.
+        names = mod.all
+        if names is None:
+            names = [
+                name
+                for name in chain(mod.contents.keys(),
+                                  mod._localNameToFullName_map.keys())
+                if not name.startswith('_')
+                ]
+
+        # Add imported names to our module namespace.
         _localNameToFullName = self.builder.current._localNameToFullName_map
-        for al in node.names:
-            fromname, asname = al.name, al.asname
-            if fromname == '*':
-                if mod is None:
-                    self.builder.warning("import * from unknown", modname)
-                    return
-                self.builder.warning("import *", modname)
-                if mod.all is not None:
-                    names = mod.all
-                else:
-                    names = (
-                        k
-                        for k in chain(mod.contents.keys(),
-                                       mod._localNameToFullName_map.keys())
-                        if not k.startswith('_')
-                        )
-                for n in names:
-                    _localNameToFullName[n] = expandName(n)
-                return
+        expandName = mod.expandName
+        for name in names:
+            _localNameToFullName[name] = expandName(name)
+
+    def _importNames(self, modname: str, names: Iterable[ast.alias]) -> None:
+        """Handle a C{from <modname> import <names>} statement."""
+
+        # Process the module we're importing from.
+        # If we're importing from a package, 'mod' will be the __init__ module.
+        mod = self.system.getProcessedModule(modname)
+        obj = self.system.objForFullName(modname)
+
+        # Are we importing from a package?
+        is_package = isinstance(obj, model.Package)
+        assert is_package or obj is mod or mod is None
+
+        # Fetch names to export.
+        current = self.builder.current
+        if isinstance(current, model.Module):
+            exports = current.all
+            if exports is None:
+                exports = []
+        else:
+            # Don't export names imported inside classes or functions.
+            exports = []
+
+        _localNameToFullName = current._localNameToFullName_map
+        for al in names:
+            orgname, asname = al.name, al.asname
             if asname is None:
-                asname = fromname
-            if isinstance(self.builder.current, model.Module) and \
-                   self.builder.current.all is not None and \
-                   asname in self.builder.current.all and \
-                   modname in self.system.allobjects:
-                mod = self.system.allobjects[modname]
-                if isinstance(mod, model.Module) and \
-                       fromname in mod.contents and \
-                       (mod.all is None or fromname not in mod.all):
-                    self.system.msg(
-                        "astbuilder",
-                        "moving %r into %r"
-                        % (mod.contents[fromname].fullName(),
-                           self.builder.current.fullName()))
-                    ob = mod.contents[fromname]
-                    ob.reparent(self.builder.current, asname)
-                    continue
-            if isinstance(
-                self.system.objForFullName(modname), model.Package):
-                self.system.getProcessedModule(modname + '.' + fromname)
-            _localNameToFullName[asname] = expandName(fromname)
+                asname = orgname
+
+            # Move re-exported objects into current module.
+            if asname in exports and mod is not None:
+                try:
+                    ob = mod.contents[orgname]
+                except KeyError:
+                    self.builder.warning("cannot find re-exported name",
+                                         f'{modname}.{orgname}')
+                else:
+                    if mod.all is None or orgname not in mod.all:
+                        self.system.msg(
+                            "astbuilder",
+                            "moving %r into %r" % (ob.fullName(), current.fullName())
+                            )
+                        ob.reparent(current, asname)
+                        continue
+
+            # If we're importing from a package, make sure imported modules
+            # are processed (getProcessedModule() ignores non-modules).
+            if is_package:
+                self.system.getProcessedModule(f'{modname}.{orgname}')
+
+            _localNameToFullName[asname] = f'{modname}.{orgname}'
 
     def visit_Import(self, node):
         """Process an import statement.
@@ -280,9 +315,10 @@ class ModuleVistor(ast.NodeVisitor):
             # It is handled by findAll(), which operates on the AST and
             # therefore doesn't need an Attribute instance.
             return
-        obj = self.builder.current.resolveName(target)
+        parent = self.builder.current
+        obj = parent.resolveName(target)
         if obj is None:
-            obj = self.builder.addAttribute(target, None)
+            obj = self.builder.addAttribute(target, None, parent)
         if isinstance(obj, model.Attribute):
             obj.kind = 'Variable'
             obj.annotation = annotation
@@ -293,31 +329,46 @@ class ModuleVistor(ast.NodeVisitor):
         if not self._handleAliasing(target, expr):
             self._handleModuleVar(target, annotation, lineno)
 
-    def _handleClassVar(self, target, annotation, lineno):
-        obj = self.builder.current.contents.get(target)
-        if not isinstance(obj, model.Attribute):
-            obj = self.builder.addAttribute(target, None)
+    def _maybeAttribute(self, cls: model.Class, name: str) -> bool:
+        """Check whether a name is a potential attribute of the given class.
+        This is used to prevent an assignment that wraps a method from
+        creating an attribute that would overwrite or shadow that method.
+
+        @return: L{True} if the name does not exist or is an existing (possibly
+            inherited) attribute, L{False} otherwise
+        """
+        obj = cls.find(name)
+        return obj is None or isinstance(obj, model.Attribute)
+
+    def _handleClassVar(self, name: str, annotation: Optional[ast.expr], lineno: int) -> None:
+        cls = self.builder.current
+        if not self._maybeAttribute(cls, name):
+            return
+        obj = cls.contents.get(name)
+        if obj is None:
+            obj = self.builder.addAttribute(name, None, cls)
         if obj.kind is None:
             obj.kind = 'Class Variable'
         obj.annotation = annotation
         obj.setLineNumber(lineno)
         self.newAttr = obj
 
-    def _handleInstanceVar(self, target, annotation, lineno):
+    def _handleInstanceVar(self, name: str, annotation: Optional[ast.expr], lineno: int) -> None:
         func = self.builder.current
         if not isinstance(func, model.Function):
             return
         cls = func.parent
         if not isinstance(cls, model.Class):
             return
-        obj = cls.contents.get(target)
+        if not self._maybeAttribute(cls, name):
+            return
+        obj = cls.contents.get(name)
         if obj is None:
-            obj = self.builder.addAttribute(target, None, cls)
-        if isinstance(obj, model.Attribute):
-            obj.kind = 'Instance Variable'
-            obj.annotation = annotation
-            obj.setLineNumber(lineno)
-            self.newAttr = obj
+            obj = self.builder.addAttribute(name, None, cls)
+        obj.kind = 'Instance Variable'
+        obj.annotation = annotation
+        obj.setLineNumber(lineno)
+        self.newAttr = obj
 
     def _handleAssignmentInClass(self, target, annotation, expr, lineno):
         if not self._handleAliasing(target, expr):
@@ -326,8 +377,7 @@ class ModuleVistor(ast.NodeVisitor):
     def _handleDocstringUpdate(self, targetNode, expr, lineno):
         def warn(msg):
             self.system.msg('ast', "%s:%d: %s" % (
-                    getattr(self.builder.currentMod, 'filepath', '<unknown>'),
-                    lineno, msg))
+                    self.builder.currentMod.description, lineno, msg))
 
         # Figure out target object.
         full_name = node2fullname(targetNode, self.builder.current)
@@ -348,7 +398,7 @@ class ModuleVistor(ast.NodeVisitor):
             warn("Unable to figure out value for __doc__ assignment, "
                  "maybe too complex")
             return
-        if not isinstance(docstring, string_types):
+        if not isinstance(docstring, str):
             warn("Ignoring value assigned to __doc__: not a string")
             return
 
@@ -407,68 +457,164 @@ class ModuleVistor(ast.NodeVisitor):
 
         self.generic_visit(node)
 
-    def visit_FunctionDef(self, node):
+    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
+        self._handleFunctionDef(node, is_async=True)
+
+    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
+        self._handleFunctionDef(node, is_async=False)
+
+    def _handleFunctionDef(self,
+            node: Union[ast.AsyncFunctionDef, ast.FunctionDef],
+            is_async: bool
+            ) -> None:
         # Ignore inner functions.
-        if isinstance(self.builder.current, model.Function):
+        parent = self.builder.current
+        if isinstance(parent, model.Function):
             return
 
         lineno = node.lineno
         if node.decorator_list:
             lineno = node.decorator_list[0].lineno
 
-        func = self.builder.pushFunction(node.name, lineno)
-        if len(node.body) > 0 and isinstance(node.body[0], ast.Expr) and isinstance(node.body[0].value, ast.Str):
-            func.setDocstring(node.body[0].value)
-        func.decorators = node.decorator_list
-        if isinstance(func.parent, model.Class) and node.decorator_list:
-            isclassmethod = False
-            isstaticmethod = False
+        docstring: Optional[ast.Str] = None
+        if len(node.body) > 0 and isinstance(node.body[0], ast.Expr) \
+                              and isinstance(node.body[0].value, ast.Str):
+            docstring = node.body[0].value
+
+        func_name = node.name
+        is_property = False
+        is_classmethod = False
+        is_staticmethod = False
+        if isinstance(parent, model.Class) and node.decorator_list:
             for d in node.decorator_list:
                 if isinstance(d, ast.Name):
-                    if d.id == 'classmethod':
-                        isclassmethod = True
-                    elif d.id == 'staticmethod':
-                        isstaticmethod = True
-            if isstaticmethod:
-                if isclassmethod:
-                    func.report('%s is both classmethod and staticmethod' % (
-                                func.fullName(),))
-                else:
-                    func.kind = 'Static Method'
-            elif isclassmethod:
-                func.kind = 'Class Method'
+                    name = d.id
+                    if name == 'property':
+                        is_property = True
+                    elif name.endswith('property') or name.endswith('Property'):
+                        is_property = True
+                    elif name == 'classmethod':
+                        is_classmethod = True
+                    elif name == 'staticmethod':
+                        is_staticmethod = True
+                elif isinstance(d, ast.Attribute):
+                    if d.attr in ('setter', 'deleter'):
+                        # Rename the setter/deleter, so it doesn't replace
+                        # the property object.
+                        func_name = f'{func_name}.{d.attr}'
+                elif isinstance(d, ast.Call):
+                    deco_name = node2fullname(d.func, parent)
+                    if deco_name is not None and (
+                            deco_name.endswith('property') or
+                            deco_name.endswith('Property')
+                            ):
+                        is_property = True
 
-        args = []
+        if is_property:
+            attr = self._handlePropertyDef(node, docstring, lineno)
+            if is_classmethod:
+                attr.report(f'{attr.fullName()} is both property and classmethod')
+            if is_staticmethod:
+                attr.report(f'{attr.fullName()} is both property and staticmethod')
+            return
 
-        for arg in node.args.args:
-            if isinstance(arg, (ast.Tuple, ast.List)):
-                args.append([x.id for x in arg.elts])
-            elif isinstance(arg, ast.Name):
-                args.append(arg.id)
+        func = self.builder.pushFunction(func_name, lineno)
+        func.is_async = is_async
+        if docstring is not None:
+            func.setDocstring(docstring)
+        func.decorators = node.decorator_list
+        if is_staticmethod:
+            if is_classmethod:
+                func.report(f'{func.fullName()} is both classmethod and staticmethod')
             else:
-                args.append(arg.arg)
+                func.kind = 'Static Method'
+        elif is_classmethod:
+            func.kind = 'Class Method'
 
-        varargname = node.args.vararg
-        if varargname and not isinstance(varargname, string_types):
-            varargname = varargname.arg
+        # Position-only arguments were introduced in Python 3.8.
+        posonlyargs: Sequence[ast.arg] = getattr(node.args, 'posonlyargs', ())
 
-        kwargname = node.args.kwarg
-        if kwargname and not isinstance(kwargname, string_types):
-            kwargname = kwargname.arg
+        num_pos_args = len(posonlyargs) + len(node.args.args)
+        defaults = node.args.defaults
+        default_offset = num_pos_args - len(defaults)
+        def get_default(index: int) -> Optional[ast.expr]:
+            assert 0 <= index < num_pos_args, index
+            index -= default_offset
+            return None if index < 0 else defaults[index]
 
-        defaults = []
+        parameters = []
+        def add_arg(name: str, kind: Any, default: Optional[ast.expr]) -> None:
+            default_val = Parameter.empty if default is None else _ValueFormatter(default)
+            parameters.append(Parameter(name, kind, default=default_val))
 
-        for default in node.args.defaults:
-            if isinstance(default, ast.Num):
-                defaults.append(str(default.n))
-            else:
-                defaults.append(astor.to_source(default).strip())
+        for index, arg in enumerate(posonlyargs):
+            add_arg(arg.arg, Parameter.POSITIONAL_ONLY, get_default(index))
 
-        func.argspec = (args, varargname, kwargname, tuple(defaults))
+        for index, arg in enumerate(node.args.args, start=len(posonlyargs)):
+            add_arg(arg.arg, Parameter.POSITIONAL_OR_KEYWORD, get_default(index))
+
+        vararg = node.args.vararg
+        if vararg is not None:
+            add_arg(vararg.arg, Parameter.VAR_POSITIONAL, None)
+
+        assert len(node.args.kwonlyargs) == len(node.args.kw_defaults)
+        for arg, default in zip(node.args.kwonlyargs, node.args.kw_defaults):
+            add_arg(arg.arg, Parameter.KEYWORD_ONLY, default)
+
+        kwarg = node.args.kwarg
+        if kwarg is not None:
+            add_arg(kwarg.arg, Parameter.VAR_KEYWORD, None)
+
+        try:
+            signature = Signature(parameters)
+        except ValueError as ex:
+            func.report(f'{func.fullName()} has invalid parameters: {ex}')
+            signature = Signature()
+
+        func.signature = signature
+        func.annotations = self._annotations_from_function(node)
         self.default(node)
         self.builder.popFunction()
 
-    def _annotation_from_attrib(self, expr, ctx):
+    def _handlePropertyDef(self,
+            node: Union[ast.AsyncFunctionDef, ast.FunctionDef],
+            docstring: Optional[ast.Str],
+            lineno: int
+            ) -> model.Attribute:
+
+        attr = self.builder.addAttribute(node.name, 'Property', self.builder.current)
+        attr.setLineNumber(lineno)
+
+        if docstring is not None:
+            attr.setDocstring(docstring)
+            assert attr.docstring is not None
+            pdoc = epydoc2stan.parse_docstring(attr, attr.docstring, attr)
+            other_fields = []
+            for field in pdoc.fields:
+                tag = field.tag()
+                if tag == 'return':
+                    if not pdoc.has_body:
+                        pdoc = field.body()
+                        # Avoid format_summary() going back to the original
+                        # empty-body docstring.
+                        attr.docstring = ''
+                elif tag == 'rtype':
+                    attr.parsed_type = field.body()
+                else:
+                    other_fields.append(field)
+            pdoc.fields = other_fields
+            attr.parsed_docstring = pdoc
+
+        if node.returns is not None:
+            attr.annotation = self._unstring_annotation(node.returns)
+        attr.decorators = node.decorator_list
+
+        return attr
+
+    def _annotation_from_attrib(self,
+            expr: ast.expr,
+            ctx: model.Documentable
+            ) -> Optional[ast.expr]:
         """Get the type of an C{attr.ib} definition.
         @param expr: The expression's AST.
         @param ctx: The context in which this expression is evaluated.
@@ -486,38 +632,139 @@ class ModuleVistor(ast.NodeVisitor):
                 return _infer_type(default)
         return None
 
-    def _unstring_annotation(self, node):
+    def _annotations_from_function(
+            self, func: Union[ast.AsyncFunctionDef, ast.FunctionDef]
+            ) -> Mapping[str, Optional[ast.expr]]:
+        """Get annotations from a function definition.
+        @param func: The function definition's AST.
+        @return: Mapping from argument name to annotation.
+            The name C{return} is used for the return type.
+            Unannotated arguments are omitted.
+        """
+        def _get_all_args() -> Iterator[ast.arg]:
+            base_args = func.args
+            # New on Python 3.8 -- handle absence gracefully
+            try:
+                yield from base_args.posonlyargs
+            except AttributeError:
+                pass
+            yield from base_args.args
+            varargs = base_args.vararg
+            if varargs:
+                yield varargs
+            yield from base_args.kwonlyargs
+            kwargs = base_args.kwarg
+            if kwargs:
+                yield kwargs
+        def _get_all_ast_annotations() -> Iterator[Tuple[str, Optional[ast.expr]]]:
+            for arg in _get_all_args():
+                yield arg.arg, arg.annotation
+            returns = func.returns
+            if returns:
+                yield 'return', returns
+        return {
+            # Include parameter names even if they're not annotated, so that
+            # we can use the key set to know which parameters exist and warn
+            # when non-existing parameters are documented.
+            name: None if value is None else self._unstring_annotation(value)
+            for name, value in _get_all_ast_annotations()
+            }
+
+    def _unstring_annotation(self, node: ast.expr) -> ast.expr:
         """Replace all strings in the given expression by parsed versions.
-        Returns the resulting node, or None if parsing failed.
+        @return: The unstringed node. If parsing fails, an error is logged
+            and the original node is returned.
         """
         try:
-            return _AnnotationStringParser().visit(node)
+            expr = _AnnotationStringParser().visit(node)
         except SyntaxError as ex:
             self.builder.currentMod.report(
-                    'syntax error in annotation: %s' % (ex,),
+                    f'syntax error in annotation: {ex}',
                     lineno_offset=node.lineno)
-            return None
+            return node
+        else:
+            assert isinstance(expr, ast.expr), expr
+            return expr
+
+
+class _ValueFormatter:
+    """Formats values stored in AST expressions.
+    Used for presenting default values of parameters.
+    """
+
+    def __init__(self, value: ast.expr):
+        self.value = value
+
+    def __repr__(self) -> str:
+        value = self.value
+        if isinstance(value, ast.Num):
+            return str(value.n)
+        if isinstance(value, ast.Str):
+            return repr(value.s)
+        if isinstance(value, ast.Constant):
+            return repr(value.value)
+        if isinstance(value, ast.UnaryOp) and isinstance(value.op, ast.USub):
+            operand = value.operand
+            if isinstance(operand, ast.Num):
+                return f'-{operand.n}'
+            if isinstance(operand, ast.Constant):
+                return f'-{operand.value}'
+        source: str = astor.to_source(value)
+        return source.strip()
 
 
 class _AnnotationStringParser(ast.NodeTransformer):
+    """Implementation of L{ModuleVistor._unstring_annotation()}.
+
+    When given an expression, the node returned by L{ast.NodeVisitor.visit()}
+    will also be an expression.
+    If any string literal contained in the original expression is either
+    invalid Python or not a singular expression, L{SyntaxError} is raised.
+    """
+
+    def _parse_string(self, value: str) -> ast.expr:
+        statements = ast.parse(value).body
+        if len(statements) != 1:
+            raise SyntaxError("expected expression, found multiple statements")
+        stmt, = statements
+        if isinstance(stmt, ast.Expr):
+            # Expression wrapped in an Expr statement.
+            expr = self.visit(stmt.value)
+            assert isinstance(expr, ast.expr), expr
+            return expr
+        else:
+            raise SyntaxError("expected expression, found statement")
+
+    def visit_Subscript(self, node: ast.Subscript) -> ast.Subscript:
+        value = self.visit(node.value)
+        if isinstance(value, ast.Name) and value.id == 'Literal':
+            # Literal[...] expression; don't unstring the arguments.
+            slice = node.slice
+        elif isinstance(value, ast.Attribute) and value.attr == 'Literal':
+            # typing.Literal[...] expression; don't unstring the arguments.
+            slice = node.slice
+        else:
+            # Other subscript; unstring the slice.
+            slice = self.visit(node.slice)
+        return ast.copy_location(ast.Subscript(value, slice, node.ctx), node)
 
     # For Python >= 3.8:
 
-    def visit_Constant(self, node):
+    def visit_Constant(self, node: ast.Constant) -> ast.expr:
         value = node.value
         if isinstance(value, str):
-            expr, = ast.parse(value).body
-            return self.visit(expr.value)
+            return ast.copy_location(self._parse_string(value), node)
         else:
-            return self.generic_visit(node)
+            const = self.generic_visit(node)
+            assert isinstance(const, ast.Constant), const
+            return const
 
     # For Python < 3.8:
 
-    def visit_Str(self, node):
-        expr, = ast.parse(node.s).body
-        return self.visit(expr.value)
+    def visit_Str(self, node: ast.Str) -> ast.expr:
+        return ast.copy_location(self._parse_string(node.s), node)
 
-def _infer_type(expr):
+def _infer_type(expr: ast.expr) -> Optional[ast.expr]:
     """Infer an expression's type.
     @param expr: The expression's AST.
     @return: A type annotation, or None if the expression has no obvious type.
@@ -528,14 +775,17 @@ def _infer_type(expr):
     except ValueError:
         return None
     else:
-        return _annotation_for_value(value)
+        ann = _annotation_for_value(value)
+        if ann is None:
+            return None
+        else:
+            return ast.fix_missing_locations(ast.copy_location(ann, expr))
 
-def _annotation_for_value(value):
+def _annotation_for_value(value: object) -> Optional[ast.expr]:
     if value is None:
         return None
     name = type(value).__name__
     if isinstance(value, (dict, list, set, tuple)):
-        name = name.capitalize()
         ann_elem = _annotation_for_elements(value)
         if isinstance(value, dict):
             ann_value = _annotation_for_elements(value.values())
@@ -544,13 +794,13 @@ def _annotation_for_value(value):
             elif ann_elem is not None:
                 ann_elem = ast.Tuple(elts=[ann_elem, ann_value])
         if ann_elem is not None:
-            if name == 'Tuple':
+            if name == 'tuple':
                 ann_elem = ast.Tuple(elts=[ann_elem, ast.Ellipsis()])
             return ast.Subscript(value=ast.Name(id=name),
                                  slice=ast.Index(value=ann_elem))
     return ast.Name(id=name)
 
-def _annotation_for_elements(sequence):
+def _annotation_for_elements(sequence: Iterable[object]) -> Optional[ast.expr]:
     names = set()
     for elem in sequence:
         ann = _annotation_for_value(elem)
@@ -567,9 +817,12 @@ def _annotation_for_elements(sequence):
         return None
 
 
-class ASTBuilder(object):
+class ASTBuilder:
     ModuleVistor = ModuleVistor
 
+    system: model.System
+    ast_cache: Dict[Path, Optional[ast.AST]]
+
     def __init__(self, system):
         self.system = system
         self.current = None
@@ -604,7 +857,7 @@ class ASTBuilder(object):
             obj.setLineNumber(lineno)
 
     def pop(self, obj):
-        assert self.current is obj, "%r is not %r"%(self.current, obj)
+        assert self.current is obj, f"{self.current!r} is not {obj!r}"
         self.current = self._stack.pop()
         if isinstance(obj, model.Module):
             self.currentMod = None
@@ -619,35 +872,36 @@ class ASTBuilder(object):
     def popFunction(self):
         self._pop(self.system.Function)
 
-    def addAttribute(self, target, kind, parent=None):
-        if parent is None:
-            parent = self.current
+    def addAttribute(self,
+            name: str, kind: Optional[str], parent: model.Documentable
+            ) -> model.Attribute:
         system = self.system
         parentMod = self.currentMod
-        attr = system.Attribute(system, target, parent)
+        attr = system.Attribute(system, name, parent)
         attr.kind = kind
         attr.parentMod = parentMod
         system.addObject(attr)
         return attr
 
-    def warning(self, type, detail):
-        self.system._warning(self.current, type, detail)
+    def warning(self, message, detail):
+        self.system._warning(self.current, message, detail)
 
     def processModuleAST(self, ast, mod):
         findAll(ast, mod)
 
         self.ModuleVistor(self, mod).visit(ast)
 
-    def parseFile(self, filePath):
-        if filePath in self.ast_cache:
-            return self.ast_cache[filePath]
+    def parseFile(self, path: Path) -> Optional[ast.AST]:
         try:
-            ast = parseFile(filePath)
-        except (SyntaxError, ValueError):
-            self.warning("cannot parse", filePath)
-            ast = None
-        self.ast_cache[filePath] = ast
-        return ast
+            return self.ast_cache[path]
+        except KeyError:
+            mod: Optional[ast.AST] = None
+            try:
+                mod = parseFile(path)
+            except (SyntaxError, ValueError):
+                self.warning("cannot parse", str(path))
+            self.ast_cache[path] = mod
+            return mod
 
 model.System.defaultBuilder = ASTBuilder
 
