@@ -1,7 +1,8 @@
 """The command-line parsing and entry point."""
 
-from __future__ import print_function
-
+from optparse import SUPPRESS_HELP, Option, OptionParser, OptionValueError, Values
+from pathlib import Path
+from typing import TYPE_CHECKING, List, Sequence, Tuple, Type, TypeVar, cast
 import datetime
 import os
 import sys
@@ -10,16 +11,31 @@ from pydoctor import model, zopeinterface, __version__
 from pydoctor.sphinx import (MAX_AGE_HELP, USER_INTERSPHINX_CACHE,
                              SphinxInventoryWriter, prepareCache)
 
+if TYPE_CHECKING:
+    from typing_extensions import NoReturn
+else:
+    NoReturn = None
+
+
 BUILDTIME_FORMAT = '%Y-%m-%d %H:%M:%S'
 
-def error(msg, *args):
+def error(msg: str, *args: object) -> NoReturn:
     if args:
         msg = msg%args
     print(msg, file=sys.stderr)
     sys.exit(1)
 
-def findClassFromDottedName(dottedname, optionname):
-    # watch out, prints a message and SystemExits on error!
+T = TypeVar('T')
+
+def findClassFromDottedName(
+        dottedname: str,
+        optionname: str,
+        base_class: Type[T]
+        ) -> Type[T]:
+    """
+    Looks up a class by full name.
+    Watch out, prints a message and SystemExits on error!
+    """
     if '.' not in dottedname:
         error("%stakes a dotted name", optionname)
     parts = dottedname.rsplit('.', 1)
@@ -28,15 +44,37 @@ def findClassFromDottedName(dottedname, optionname):
     except ImportError:
         error("could not import module %s", parts[0])
     try:
-        return getattr(mod, parts[1])
+        cls = getattr(mod, parts[1])
     except AttributeError:
         error("did not find %s in module %s", parts[1], parts[0])
+    if not issubclass(cls, base_class):
+        error("%s is not a subclass of %s", cls, base_class)
+    return cast(Type[T], cls)
 
 MAKE_HTML_DEFAULT = object()
 
-def getparser():
-    from optparse import OptionParser
-    parser = OptionParser(version=__version__.public())
+def parse_path(option: Option, opt: str, value: str) -> Path:
+    """Parse a path value given to an option to a L{Path} object.
+    The path is converted to an absolute path, as required by
+    L{System.setSourceHref()}.
+    The path does not need to exist.
+    """
+    try:
+        # We explicitly make the path relative to the current working dir
+        # because on Windows resolve() does not produce an absolute path
+        # when operating on a non-existing path.
+        return Path(Path.cwd(), value).resolve()
+    except Exception as ex:
+        raise OptionValueError(f"{opt}: invalid path: {ex}")
+
+class CustomOption(Option):
+    TYPES = Option.TYPES + ("path",)
+    TYPE_CHECKER = dict(Option.TYPE_CHECKER, path=parse_path)
+
+def getparser() -> OptionParser:
+    parser = OptionParser(
+        option_class=CustomOption, version=__version__.public(),
+        usage="usage: %prog [options] SOURCEPATH...")
     parser.add_option(
         '-c', '--config', dest='configfile',
         help=("Use config from this file (any command line"
@@ -51,10 +89,9 @@ def getparser():
         '--project-url', dest='projecturl',
         help=("The project url, appears in the html if given."))
     parser.add_option(
-        '--project-base-dir', dest='projectbasedirectory',
-        help=("Absolute path to the base directory of the "
-              "project.  Source links will be computed based "
-              "on this value."))
+        '--project-base-dir', dest='projectbasedirectory', type='path',
+        help=("Path to the base directory of the project.  Source links "
+              "will be computed based on this value."))
     parser.add_option(
         '--testing', dest='testing', action='store_true',
         help=("Don't complain if the run doesn't have any effects."))
@@ -69,26 +106,14 @@ def getparser():
         default=False, help=("Produce (only) the objects.inv intersphinx file."))
     parser.add_option(
         '--add-package', action='append', dest='packages',
-        metavar='PACKAGEDIR', default=[],
-        help=("Add a package to the system.  Can be repeated "
-              "to add more than one package."))
+        metavar='PACKAGEDIR', default=[], help=SUPPRESS_HELP)
     parser.add_option(
         '--add-module', action='append', dest='modules',
-        metavar='MODULE', default=[],
-        help=("Add a module to the system.  Can be repeated."))
+        metavar='MODULE', default=[], help=SUPPRESS_HELP)
     parser.add_option(
         '--prepend-package', action='store', dest='prependedpackage',
         help=("Pretend that all packages are within this one.  "
               "Can be used to document part of a package."))
-    parser.add_option(
-        '--abbreviate-specialcase', action='store',
-        dest='abbrevmapping', default='',
-        help=("This is a comma seperated list of key=value pairs.  "
-              "Where any key corresponds to a module name and value is "
-              "the desired abbreviation.  This can be used to resolve "
-              "conflicts with abbreviation where you have two or more "
-              "modules that start with the same letter.  Example: "
-              "twistedcaldav=tcd."))
     parser.add_option(
         '--docformat', dest='docformat', action='store', default='epytext',
         help=("Which epydoc-supported format docstrings are assumed "
@@ -103,11 +128,7 @@ def getparser():
         help=("Only generate the summary pages."))
     parser.add_option(
         '--html-write-function-pages', dest='htmlfunctionpages',
-        default=False, action='store_true',
-        help=("Make individual HTML files for every function and "
-              "method. They're not linked to in any pydoctor-"
-              "generated HTML, but they can be useful for third-party "
-              "linking."))
+        default=False, action='store_true', help=SUPPRESS_HELP)
     parser.add_option(
         '--html-output', dest='htmloutput', default='apidocs',
         help=("Directory to save HTML files to (default 'apidocs')"))
@@ -123,6 +144,10 @@ def getparser():
         '--buildtime', dest='buildtime',
         help=("Use the specified build time over the current time. "
               "Format: %s" % BUILDTIME_FORMAT))
+    parser.add_option(
+        '-W', '--warnings-as-errors', action='store_true',
+        dest='warnings_as_errors', default=False,
+        help=("Return exit code 3 on warnings."))
     parser.add_option(
         '-v', '--verbose', action='count', dest='verbosity',
         default=0,
@@ -131,7 +156,8 @@ def getparser():
         '-q', '--quiet', action='count', dest='quietness',
         default=0,
         help=("Be quieter."))
-    def verbose_about_callback(option, opt_str, value, parser):
+    def verbose_about_callback(option: Option, opt_str: str, value: str, parser: OptionParser) -> None:
+        assert parser.values is not None
         d = parser.values.verbosity_details
         d[value] = d.get(value, 0) + 1
     parser.add_option(
@@ -147,15 +173,22 @@ def getparser():
         '--intersphinx', action='append', dest='intersphinx',
         metavar='URL_TO_OBJECTS.INV', default=[],
         help=(
-            "Use Sphinx objects inventory to generate links to external"
-            "documetation. Can be repeated."))
+            "Use Sphinx objects inventory to generate links to external "
+            "documentation. Can be repeated."))
 
     parser.add_option(
         '--enable-intersphinx-cache',
-        dest='enable_intersphinx_cache',
+        dest='enable_intersphinx_cache_deprecated',
         action='store_true',
         default=False,
-        help="Enable Intersphinx cache."
+        help=SUPPRESS_HELP
+    )
+    parser.add_option(
+        '--disable-intersphinx-cache',
+        dest='enable_intersphinx_cache',
+        action='store_false',
+        default=True,
+        help="Disable Intersphinx cache."
     )
     parser.add_option(
         '--intersphinx-cache-path',
@@ -180,9 +213,9 @@ def getparser():
 
     return parser
 
-def readConfigFile(options):
+def readConfigFile(options: Values) -> None:
     # this is all a bit horrible.  rethink, then rewrite!
-    for i, line in enumerate(open(options.configfile, 'rU')):
+    for i, line in enumerate(open(options.configfile)):
         line = line.strip()
         if not line or line.startswith('#'):
             continue
@@ -206,13 +239,13 @@ def readConfigFile(options):
             if not isinstance(pre_v, list):
                 setattr(options, k, v)
 
-def parse_args(args):
+def parse_args(args: Sequence[str]) -> Tuple[Values, List[str]]:
     parser = getparser()
     options, args = parser.parse_args(args)
     options.verbosity -= options.quietness
     return options, args
 
-def main(args=sys.argv[1:]):
+def main(args: Sequence[str] = sys.argv[1:]) -> int:
     options, args = parse_args(args)
 
     exitcode = 0
@@ -220,6 +253,24 @@ def main(args=sys.argv[1:]):
     if options.configfile:
         readConfigFile(options)
 
+    if options.enable_intersphinx_cache_deprecated:
+        print("The --enable-intersphinx-cache option is deprecated; "
+              "the cache is now enabled by default.",
+              file=sys.stderr, flush=True)
+    if options.modules:
+        print("The --add-module option is deprecated; "
+              "pass modules as positional arguments instead.",
+              file=sys.stderr, flush=True)
+    if options.packages:
+        print("The --add-package option is deprecated; "
+              "pass packages as positional arguments instead.",
+              file=sys.stderr, flush=True)
+    if options.htmlfunctionpages:
+        print("The --html-write-function-pages option is deprecated; "
+              "use the generated Intersphinx inventory (objects.inv) "
+              "for deep-linking your documentation.",
+              file=sys.stderr, flush=True)
+
     cache = prepareCache(clearCache=options.clear_intersphinx_cache,
                          enableCache=options.enable_intersphinx_cache,
                          cachePath=options.intersphinx_cache_path,
@@ -228,11 +279,8 @@ def main(args=sys.argv[1:]):
     try:
         # step 1: make/find the system
         if options.systemclass:
-            systemclass = findClassFromDottedName(options.systemclass,
-                                                  '--system-class')
-            if not issubclass(systemclass, model.System):
-                msg = "%s is not a subclass of model.System"
-                error(msg, systemclass)
+            systemclass = findClassFromDottedName(
+                options.systemclass, '--system-class', model.System)
         else:
             systemclass = zopeinterface.ZopeInterfaceSystem
 
@@ -245,11 +293,6 @@ def main(args=sys.argv[1:]):
                       "when using --html-viewsource-base")
             system.sourcebase = options.htmlsourcebase
 
-        if options.abbrevmapping:
-            for thing in options.abbrevmapping.split(','):
-                k, v = thing.split('=')
-                system.abbrevmapping[k] = v
-
         # step 1.5: check that we're actually going to accomplish something here
         args = list(args) + options.modules + options.packages
 
@@ -265,7 +308,7 @@ def main(args=sys.argv[1:]):
             system.buildtime = datetime.datetime.utcfromtimestamp(
                 int(os.environ['SOURCE_DATE_EPOCH']))
         except ValueError as e:
-            error(e)
+            error(str(e))
         except KeyError:
             pass
 
@@ -274,7 +317,7 @@ def main(args=sys.argv[1:]):
                 system.buildtime = datetime.datetime.strptime(
                     options.buildtime, BUILDTIME_FORMAT)
             except ValueError as e:
-                error(e)
+                error(str(e))
 
         # step 2: add any packages and modules
 
@@ -287,44 +330,46 @@ def main(args=sys.argv[1:]):
                     system.addObject(prependedpackage)
                     initmodule = system.Module(system, '__init__', prependedpackage)
                     system.addObject(initmodule)
+            added_paths = set()
             for path in args:
                 path = os.path.abspath(path)
-                if path in system.packages:
+                if path in added_paths:
                     continue
                 if os.path.isdir(path):
                     system.msg('addPackage', 'adding directory ' + path)
                     system.addPackage(path, prependedpackage)
-                else:
+                elif os.path.isfile(path):
                     system.msg('addModuleFromPath', 'adding module ' + path)
                     system.addModuleFromPath(prependedpackage, path)
-                system.packages.append(path)
+                elif os.path.exists(path):
+                    error(f"Source path is neither file nor directory: {path}")
+                else:
+                    error(f"Source path does not exist: {path}")
+                added_paths.add(path)
+        else:
+            error("No source paths given.")
 
         # step 3: move the system to the desired state
 
-        if not system.packages:
-            error("The system does not contain any code, did you "
-                  "forget an --add-package?")
-
-        system.process()
-
         if system.options.projectname is None:
-            name = '/'.join([ro.name for ro in system.rootobjects])
-            system.msg(
-                'warning',
-                'WARNING: guessing '+name+' for project name', thresh=-1)
+            name = '/'.join(ro.name for ro in system.rootobjects)
+            system.msg('warning', f"Guessing '{name}' for project name.", thresh=0)
             system.projectname = name
         else:
             system.projectname = system.options.projectname
 
+        system.process()
+
         # step 4: make html, if desired
 
         if options.makehtml:
             options.makeintersphinx = True
+            from pydoctor import templatewriter
             if options.htmlwriter:
                 writerclass = findClassFromDottedName(
-                    options.htmlwriter, '--html-writer')
+                    options.htmlwriter, '--html-writer',
+                    templatewriter.TemplateWriter)
             else:
-                from pydoctor import templatewriter
                 writerclass = templatewriter.TemplateWriter
 
             system.msg('html', 'writing html to %s using %s.%s'%(
@@ -332,7 +377,6 @@ def main(args=sys.argv[1:]):
                 writerclass.__name__))
 
             writer = writerclass(options.htmloutput)
-            writer.system = system
             writer.prepOutputDirectory()
 
             if options.htmlsubjects:
@@ -347,14 +391,18 @@ def main(args=sys.argv[1:]):
                 subjects = system.rootobjects
             writer.writeIndividualFiles(subjects, options.htmlfunctionpages)
             if system.docstring_syntax_errors:
-                def p(msg):
-                    system.msg(('epytext', 'epytext-summary'), msg, thresh=-1, topthresh=1)
+                def p(msg: str) -> None:
+                    system.msg('docstring-summary', msg, thresh=-1, topthresh=1)
                 p("these %s objects' docstrings contain syntax errors:"
                   %(len(system.docstring_syntax_errors),))
                 exitcode = 2
                 for fn in sorted(system.docstring_syntax_errors):
                     p('    '+fn)
 
+        if system.violations and options.warnings_as_errors:
+            # Update exit code if the run has produced warnings.
+            exitcode = 3
+
         if options.makeintersphinx:
             if not options.makehtml:
                 subjects = system.rootobjects
