@@ -6,20 +6,31 @@ system being documented.  An instance of L{System} represents the whole system
 being documented -- a System is a bad of Documentables, in some sense.
 """
 
-from __future__ import print_function, unicode_literals
-
+import ast
 import datetime
-import imp
+import importlib
 import inspect
 import os
 import platform
 import sys
 import types
-from collections import OrderedDict
 from enum import Enum
+from inspect import Signature
+from optparse import Values
+from pathlib import Path
+from typing import (
+    TYPE_CHECKING, Any, Collection, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Set, Tuple, Type, TypeVar, Union
+)
+from urllib.parse import quote
+
+from pydoctor.epydoc.markup import ParsedDocstring
+from pydoctor.sphinx import CacheT, SphinxInventory
+
+if TYPE_CHECKING:
+    from pydoctor.astbuilder import ASTBuilder
+else:
+    ASTBuilder = object
 
-from pydoctor.sphinx import SphinxInventory
-from six.moves import builtins
 
 # originally when I started to write pydoctor I had this idea of a big
 # tree of Documentables arranged in an almost arbitrary tree.
@@ -49,7 +60,26 @@ class DocLocation(Enum):
     #UNDER_PARENT_DOCSTRING = 3
 
 
-class Documentable(object):
+class ProcessingState(Enum):
+    UNPROCESSED = 0
+    PROCESSING = 1
+    PROCESSED = 2
+
+
+class PrivacyClass(Enum):
+    """L{Enum} containing values indicating how private an object should be.
+
+    @cvar HIDDEN: Don't show the object at all.
+    @cvar PRIVATE: Show, but de-emphasize the object.
+    @cvar VISIBLE: Show the object as normal.
+    """
+
+    HIDDEN = 0
+    PRIVATE = 1
+    VISIBLE = 2
+
+
+class Documentable:
     """An object that can be documented.
 
     The interface is a bit ridiculously wide.
@@ -59,37 +89,59 @@ class Documentable(object):
     @ivar parent: ...
     @ivar parentMod: ...
     @ivar name: ...
-    @ivar documentation_location: ...
     @ivar sourceHref: ...
     @ivar kind: ...
     """
-    documentation_location = DocLocation.OWN_PAGE
-    docstring = None
+    docstring: Optional[str] = None
+    parsed_docstring: Optional[ParsedDocstring] = None
+    parsed_type: Optional[ParsedDocstring] = None
     docstring_lineno = 0
     linenumber = 0
-    sourceHref = None
+    sourceHref: Optional[str] = None
+    kind: Optional[str]
 
     @property
-    def css_class(self):
+    def documentation_location(self) -> DocLocation:
+        """Page location where we are documented.
+        The default implementation returns L{DocLocation.OWN_PAGE}.
+        """
+        return DocLocation.OWN_PAGE
+
+    @property
+    def css_class(self) -> str:
         """A short, lower case description for use as a CSS class in HTML."""
-        class_ = self.kind.lower().replace(' ', '')
+        kind = self.kind
+        assert kind is not None # if kind is None, object is invisible
+        class_ = kind.lower().replace(' ', '')
         if self.privacyClass is PrivacyClass.PRIVATE:
             class_ += ' private'
         return class_
 
-    def __init__(self, system, name, parent=None):
+    @property
+    def doctarget(self) -> 'Documentable':
+        return self
+
+    def __init__(
+            self, system: 'System', name: str,
+            parent: Optional['Documentable'] = None,
+            source_path: Optional[Path] = None
+            ):
+        if not isinstance(self, Package):
+            if source_path is None and parent is not None:
+                source_path = parent.source_path # type: ignore[has-type]
         self.system = system
         self.name = name
         self.parent = parent
-        self.parentMod = None
+        self.parentMod: Optional[Module] = None
+        self.source_path = source_path
         self.setup()
-        if not isinstance(self, Package):
-            self.doctarget = self
 
-    def setup(self):
-        self.contents = OrderedDict()
+    def setup(self) -> None:
+        # TODO: The actual value type is Documentable, but using that
+        #       requires a boatload of changes.
+        self.contents: Dict[str, Any] = {}
 
-    def setDocstring(self, node):
+    def setDocstring(self, node: ast.Str) -> None:
         doc = node.s
         lineno = node.lineno
         if _string_lineno_is_end:
@@ -110,16 +162,44 @@ class Documentable(object):
         self.docstring = inspect.cleandoc(doc)
         self.docstring_lineno = lineno
 
-    def setLineNumber(self, lineno):
+    def setLineNumber(self, lineno: int) -> None:
         if not self.linenumber:
             self.linenumber = lineno
             parentMod = self.parentMod
             if parentMod is not None:
                 parentSourceHref = parentMod.sourceHref
                 if parentSourceHref:
-                    self.sourceHref = '%s#L%d' % (parentSourceHref, lineno)
+                    self.sourceHref = f'{parentSourceHref}#L{lineno:d}'
 
-    def fullName(self):
+    @property
+    def description(self) -> str:
+        """A string describing our source location to the user.
+
+        If this module's code was read from a file, this returns
+        its file path. In other cases, such as during unit testing,
+        the full module name is returned.
+        """
+        source_path = self.source_path
+        return self.module.fullName() if source_path is None else str(source_path)
+
+    @property
+    def url(self) -> str:
+        """Relative URL at which the documentation for this Documentable
+        can be found.
+        """
+        location = self.documentation_location
+        if location is DocLocation.OWN_PAGE:
+            return f'{quote(self.fullName())}.html'
+        elif location is DocLocation.PARENT_PAGE:
+            parent = self.parent
+            if isinstance(parent, Module) and parent.name == '__init__':
+                parent = parent.parent
+            assert parent is not None
+            return f'{quote(parent.fullName())}.html#{quote(self.name)}'
+        else:
+            assert False, location
+
+    def fullName(self) -> str:
         parent = self.parent
         if parent is not None:
             if (parent.parent and isinstance(parent.parent, Package)
@@ -132,10 +212,10 @@ class Documentable(object):
             prefix = ''
         return prefix + self.name
 
-    def __repr__(self):
-        return "%s %r"%(self.__class__.__name__, self.fullName())
+    def __repr__(self) -> str:
+        return f"{self.__class__.__name__} {self.fullName()!r}"
 
-    def docsources(self):
+    def docsources(self) -> Iterator['Documentable']:
         """Objects that can be considered as a source of documentation.
 
         The motivating example for having multiple sources is looking at a
@@ -145,13 +225,14 @@ class Documentable(object):
         yield self
 
 
-    def reparent(self, new_parent, new_name):
+    def reparent(self, new_parent: 'Module', new_name: str) -> None:
         # this code attempts to preserve "rather a lot" of
         # invariants assumed by various bits of pydoctor
         # and that are of course not written down anywhere
         # :/
         self._handle_reparenting_pre()
         old_parent = self.parent
+        assert isinstance(old_parent, Module)
         old_name = self.name
         self.parent = self.parentMod = new_parent
         self.name = new_name
@@ -161,20 +242,20 @@ class Documentable(object):
         new_parent.contents[new_name] = self
         self._handle_reparenting_post()
 
-    def _handle_reparenting_pre(self):
+    def _handle_reparenting_pre(self) -> None:
         del self.system.allobjects[self.fullName()]
         for o in self.contents.values():
             o._handle_reparenting_pre()
 
-    def _handle_reparenting_post(self):
+    def _handle_reparenting_post(self) -> None:
         self.system.allobjects[self.fullName()] = self
         for o in self.contents.values():
             o._handle_reparenting_post()
 
-    def _localNameToFullName(self, name):
+    def _localNameToFullName(self, name: str) -> str:
         raise NotImplementedError(self._localNameToFullName)
 
-    def expandName(self, name):
+    def expandName(self, name: str) -> str:
         """Return a fully qualified name for the possibly-dotted `name`.
 
         To explain what this means, consider the following modules:
@@ -182,7 +263,7 @@ class Documentable(object):
         mod1.py::
 
             from external_location import External
-            class Local(object):
+            class Local:
                 pass
 
         mod2.py::
@@ -196,38 +277,44 @@ class Documentable(object):
         "external_location.External" and expandName("renamed_mod.Local")
         should be "mod1.Local". """
         parts = name.split('.')
-        obj = self
+        obj: Documentable = self
         for i, p in enumerate(parts):
             full_name = obj._localNameToFullName(p)
-            obj = self.system.objForFullName(full_name)
-            if obj is None:
+            nxt = self.system.objForFullName(full_name)
+            if nxt is None:
                 break
+            obj = nxt
         remaning = parts[i+1:]
         return '.'.join([full_name] + remaning)
 
-    def resolveName(self, name):
+    def resolveName(self, name: str) -> Optional['Documentable']:
         """Return the object named by "name" (using Python's lookup rules) in
         this context, if any is known to pydoctor."""
         return self.system.objForFullName(self.expandName(name))
 
     @property
-    def privacyClass(self):
-        """How visible this object should be.
-
-        @rtype: a member of the L{PrivacyClass} class/enum.
-        """
+    def privacyClass(self) -> PrivacyClass:
+        """How visible this object should be."""
         return self.system.privacyClass(self)
 
     @property
-    def isVisible(self):
-        """Is this object is so private as to be not shown at all?
+    def isVisible(self) -> bool:
+        """Is this object so private as to be not shown at all?
 
         This is just a simple helper which defers to self.privacyClass.
         """
         return self.privacyClass is not PrivacyClass.HIDDEN
 
     @property
-    def module(self):
+    def isPrivate(self) -> bool:
+        """Is this object considered private API?
+
+        This is just a simple helper which defers to self.privacyClass.
+        """
+        return self.privacyClass is not PrivacyClass.VISIBLE
+
+    @property
+    def module(self) -> 'Module':
         """This object's L{Module}.
 
         For modules, this returns the object itself, otherwise
@@ -237,10 +324,11 @@ class Documentable(object):
         assert parentMod is not None
         return parentMod
 
-    def report(self, descr, section='parsing', lineno_offset=0):
+    def report(self, descr: str, section: str = 'parsing', lineno_offset: int = 0) -> None:
         """Log an error or warning about this documentable object."""
 
-        if section == 'docstring':
+        linenumber: object
+        if section in ('docstring', 'resolve_identifier_xref'):
             linenumber = self.docstring_lineno
         else:
             linenumber = self.linenumber
@@ -253,108 +341,147 @@ class Documentable(object):
 
         self.system.msg(
             section,
-            '%s:%s: %s' % (self.module.description, linenumber, descr),
+            f'{self.description}:{linenumber}: {descr}',
             thresh=-1)
 
 
 class Package(Documentable):
     kind = "Package"
-    def docsources(self):
+    def docsources(self) -> Iterator[Documentable]:
         yield self.contents['__init__']
     @property
-    def doctarget(self):
-        return self.contents['__init__']
+    def doctarget(self) -> Documentable:
+        return self.contents['__init__'] # type: ignore[no-any-return]
     @property
-    def module(self):
-        return self.contents['__init__']
+    def module(self) -> 'Module':
+        return self.contents['__init__'] # type: ignore[no-any-return]
     @property
-    def state(self):
-        return self.contents['__init__'].state
+    def state(self) -> ProcessingState:
+        return self.contents['__init__'].state # type: ignore[no-any-return]
 
-    def _localNameToFullName(self, name):
+    def _localNameToFullName(self, name: str) -> str:
         if name in self.contents:
-            o = self.contents[name]
+            o: Documentable = self.contents[name]
             return o.fullName()
         else:
-            return self.contents['__init__']._localNameToFullName(name)
-
-
-class ProcessingState(Enum):
-    UNPROCESSED = 0
-    PROCESSING = 1
-    PROCESSED = 2
+            return self.module._localNameToFullName(name)
 
 
 class CanContainImportsDocumentable(Documentable):
-    def setup(self):
-        super(CanContainImportsDocumentable, self).setup()
-        self._localNameToFullName_map = {}
+    def setup(self) -> None:
+        super().setup()
+        self._localNameToFullName_map: Dict[str, str] = {}
 
 
 class Module(CanContainImportsDocumentable):
     kind = "Module"
     state = ProcessingState.UNPROCESSED
 
-    def setup(self):
-        super(Module, self).setup()
-        self.all = None
+    @property
+    def documentation_location(self) -> DocLocation:
+        if self.name == '__init__':
+            return DocLocation.PARENT_PAGE
+        else:
+            return DocLocation.OWN_PAGE
 
-    def _localNameToFullName(self, name):
+    @property
+    def privacyClass(self) -> PrivacyClass:
+        if self.name == '__main__':
+            return PrivacyClass.PRIVATE
+        else:
+            return super().privacyClass
+
+    def setup(self) -> None:
+        super().setup()
+
+        self.all: Optional[Collection[str]] = None
+        """Names listed in the C{__all__} variable of this module.
+
+        These names are considered to be exported by the module,
+        both for the purpose of C{from <module> import *} and
+        for the purpose of publishing names from private modules.
+
+        If no C{__all__} variable was found in the module, or its
+        contents could not be parsed, this is L{None}.
+        """
+
+    def _localNameToFullName(self, name: str) -> str:
         if name in self.contents:
-            o = self.contents[name]
+            o: Documentable = self.contents[name]
             return o.fullName()
         elif name in self._localNameToFullName_map:
             return self._localNameToFullName_map[name]
-        elif name in builtins.__dict__:
-            return name
         else:
             return name
 
     @property
-    def module(self):
+    def module(self) -> 'Module':
         return self
 
-    @property
-    def description(self):
-        """A string describing this module to the user.
-
-        If this module's code was read from a file, this returns
-        its file path. In other cases, such as during unit testing,
-        the module's full name is returned.
-        """
-        filepath = getattr(self, 'filepath', None)
-        return self.fullName() if filepath is None else filepath
-
 
 class Class(CanContainImportsDocumentable):
     kind = "Class"
-    def setup(self):
-        super(Class, self).setup()
-        self.rawbases = []
-        self.subclasses = []
+    parent: CanContainImportsDocumentable
+    bases: List[str]
+    baseobjects: List[Optional['Class']]
 
-    def allbases(self, include_self=False):
+    def setup(self) -> None:
+        super().setup()
+        self.rawbases: List[Class] = []
+        self.subclasses: List[Class] = []
+
+    def allbases(self, include_self: bool = False) -> Iterator['Class']:
         if include_self:
             yield self
         for b in self.baseobjects:
-            if b is None:
-                continue
-            for b2 in b.allbases(True):
-                yield b2
-    def _localNameToFullName(self, name):
+            if b is not None:
+                yield from b.allbases(True)
+
+    def find(self, name: str) -> Optional[Documentable]:
+        """Look up a name in this class and its base classes.
+
+        @return: the object with the given name, or L{None} if there isn't one
+        """
+        for base in self.allbases(include_self=True):
+            obj: Optional[Documentable] = base.contents.get(name)
+            if obj is not None:
+                return obj
+        return None
+
+    def _localNameToFullName(self, name: str) -> str:
         if name in self.contents:
-            o = self.contents[name]
+            o: Documentable = self.contents[name]
             return o.fullName()
         elif name in self._localNameToFullName_map:
             return self._localNameToFullName_map[name]
         else:
             return self.parent._localNameToFullName(name)
 
+    @property
+    def constructor_params(self) -> Mapping[str, Optional[ast.expr]]:
+        """A mapping of constructor parameter names to their type annotation.
+        If a parameter is not annotated, its value is L{None}.
+        """
+
+        # We assume that the constructor parameters are the same as the
+        # __init__() parameters. This is incorrect if __new__() or the class
+        # call have different parameters.
+        init = self.find('__init__')
+        if isinstance(init, Function):
+            return init.annotations
+        else:
+            return {}
+
 
 class Inheritable(Documentable):
-    documentation_location = DocLocation.PARENT_PAGE
 
-    def docsources(self):
+    parent: CanContainImportsDocumentable
+
+    @property
+    def documentation_location(self) -> DocLocation:
+        return DocLocation.PARENT_PAGE
+
+    def docsources(self) -> Iterator[Documentable]:
         yield self
         if not isinstance(self.parent, Class):
             return
@@ -362,36 +489,34 @@ class Inheritable(Documentable):
             if self.name in b.contents:
                 yield b.contents[self.name]
 
-    def _localNameToFullName(self, name):
+    def _localNameToFullName(self, name: str) -> str:
         return self.parent._localNameToFullName(name)
 
 class Function(Inheritable):
     kind = "Function"
+    is_async: bool
+    annotations: Mapping[str, Optional[ast.expr]]
+    decorators: Optional[Sequence[ast.expr]]
+    signature: Signature
 
-    def setup(self):
-        super(Function, self).setup()
+    def setup(self) -> None:
+        super().setup()
         if isinstance(self.parent, Class):
             self.kind = "Method"
 
 class Attribute(Inheritable):
-    kind = "Attribute"
+    kind: Optional[str] = "Attribute"
+    annotation: Optional[ast.expr]
+    decorators: Optional[Sequence[ast.expr]] = None
 
 
-class PrivacyClass(Enum):
-    """L{Enum} containing values indicating how private an object should be.
+# Work around the attributes of the same name within the System class.
+_ModuleT = Module
+_PackageT = Package
 
-    @cvar HIDDEN: Don't show the object at all.
-    @cvar PRIVATE: Show, but de-emphasize the object.
-    @cvar VISIBLE: Show the object as normal.
-    """
+T = TypeVar('T')
 
-    HIDDEN = 0
-    PRIVATE = 1
-    VISIBLE = 2
-
-
-
-class System(object):
+class System:
     """A collection of related documentable objects.
 
     PyDoctor documents collections of objects, often the contents of a
@@ -403,15 +528,21 @@ class System(object):
     Package = Package
     Function = Function
     Attribute = Attribute
-    # not done here for circularity reasons:
+    # Not assigned here for circularity reasons:
     #defaultBuilder = astbuilder.ASTBuilder
-    sourcebase = None
+    defaultBuilder: Type[ASTBuilder]
+    sourcebase: Optional[str] = None
 
-    def __init__(self, options=None):
-        self.allobjects = OrderedDict()
-        self.rootobjects = []
-        self.warnings = {}
-        self.packages = []
+    def __init__(self, options: Optional[Values] = None):
+        self.allobjects: Dict[str, Documentable] = {}
+        self.rootobjects: List[Documentable] = []
+
+        self.violations = 0
+        """The number of docstring problems found.
+        This is used to determine whether to fail the build when using
+        the --warnings-as-errors option, so it should only be increased
+        for problems that the user can fix.
+        """
 
         if options:
             self.options = options
@@ -420,47 +551,72 @@ class System(object):
             self.options, _ = parse_args([])
             self.options.verbosity = 3
 
-        self.abbrevmapping = {}
         self.projectname = 'my project'
 
-        self.docstring_syntax_errors = set()
+        self.docstring_syntax_errors: Set[str] = set()
         """FullNames of objects for which the docstring failed to parse."""
 
         self.verboselevel = 0
         self.needsnl = False
-        self.once_msgs = set()
-        self.unprocessed_modules = set()
+        self.once_msgs: Set[Tuple[str, str]] = set()
+        self.unprocessed_modules: Set[Module] = set()
         self.module_count = 0
-        self.processing_modules = []
+        self.processing_modules: List[str] = []
         self.buildtime = datetime.datetime.now()
         self.intersphinx = SphinxInventory(logger=self.msg)
 
-    def verbosity(self, section=None):
+    @property
+    def root_names(self) -> Collection[str]:
+        """The top-level package/module names in this system."""
+        return {
+            obj.name
+            for obj in self.rootobjects
+            if isinstance(obj, (Module, Package))
+            }
+
+    def verbosity(self, section: Union[str, Iterable[str]]) -> int:
         if isinstance(section, str):
             section = (section,)
-        delta = max([self.options.verbosity_details.get(sect, 0) for sect in section])
-        return self.options.verbosity + delta
+        delta: int = max(self.options.verbosity_details.get(sect, 0)
+                         for sect in section)
+        base: int = self.options.verbosity
+        return base + delta
 
-    def progress(self, section, i, n, msg):
+    def progress(self, section: str, i: int, n: Optional[int], msg: str) -> None:
         if n is None:
-            i = str(i)
+            d = str(i)
         else:
-            i = '%s/%s'%(i,n)
+            d = f'{i}/{n}'
         if self.verbosity(section) == 0 and sys.stdout.isatty():
-            print('\r'+i, msg, end='')
+            print('\r'+d, msg, end='')
             sys.stdout.flush()
-            if i == n:
+            if d == n:
                 self.needsnl = False
                 print()
             else:
                 self.needsnl = True
 
-    def msg(self, section, msg, thresh=0, topthresh=100, nonl=False, wantsnl=True, once=False):
+    def msg(self,
+            section: str,
+            msg: str,
+            thresh: int = 0,
+            topthresh: int = 100,
+            nonl: bool = False,
+            wantsnl: bool = True,
+            once: bool = False
+            ) -> None:
         if once:
             if (section, msg) in self.once_msgs:
                 return
             else:
                 self.once_msgs.add((section, msg))
+
+        if thresh < 0:
+            # Apidoc build messages are generated using negative threshold
+            # and we have separate reporting for them,
+            # on top of the logging system.
+            self.violations += 1
+
         if thresh <= self.verbosity(section) <= topthresh:
             if self.needsnl and wantsnl:
                 print()
@@ -472,25 +628,28 @@ class System(object):
                 self.needsnl = False
                 print('')
 
-    def objForFullName(self, fullName):
+    def objForFullName(self, fullName: str) -> Optional[Documentable]:
         return self.allobjects.get(fullName)
 
-    def _warning(self, current, type, detail):
+    def _warning(self,
+            current: Optional[Documentable],
+            message: str,
+            detail: str
+            ) -> None:
         if current is not None:
             fn = current.fullName()
         else:
             fn = '<None>'
         if self.options.verbosity > 0:
-            print(fn, type, detail)
-        self.warnings.setdefault(type, []).append((fn, detail))
+            print(fn, message, detail)
 
-    def objectsOfType(self, cls):
+    def objectsOfType(self, cls: Type[T]) -> Iterator[T]:
         """Iterate over all instances of C{cls} present in the system. """
         for o in self.allobjects.values():
             if isinstance(o, cls):
                 yield o
 
-    def privacyClass(self, ob):
+    def privacyClass(self, ob: Documentable) -> PrivacyClass:
         if ob.kind is None:
             return PrivacyClass.HIDDEN
         if ob.name.startswith('_') and \
@@ -498,17 +657,17 @@ class System(object):
             return PrivacyClass.PRIVATE
         return PrivacyClass.VISIBLE
 
-    def addObject(self, obj):
+    def addObject(self, obj: Documentable) -> None:
         """Add C{object} to the system."""
-        fullName = obj.fullName()
-        if obj.parent and obj.parent.fullName() != fullName:
+
+        if obj.parent:
             obj.parent.contents[obj.name] = obj
         else:
             self.rootobjects.append(obj)
-        if fullName in self.allobjects:
+
+        first = self.allobjects.setdefault(obj.fullName(), obj)
+        if obj is not first:
             self.handleDuplicate(obj)
-        else:
-            self.allobjects[fullName] = obj
 
     # if we assume:
     #
@@ -517,8 +676,7 @@ class System(object):
     # - that http://divmod.org/trac/browser/trunk is the trac URL to the
     #   above directory
     #
-    # - that ~/src/Divmod/Nevow/nevow is passed to pydoctor as an
-    #   "--add-package" argument
+    # - that ~/src/Divmod/Nevow/nevow is passed to pydoctor as an argument
     #
     # we want to work out the sourceHref for nevow.flat.ten.  the answer
     # is http://divmod.org/trac/browser/trunk/Nevow/nevow/flat/ten.py.
@@ -530,42 +688,54 @@ class System(object):
     #  http://divmod.org/trac/browser/trunk
     #                          ~/src/Divmod/Nevow/nevow/flat/ten.py
 
-    def setSourceHref(self, mod):
+    def setSourceHref(self, mod: _ModuleT, source_path: Path) -> None:
         if self.sourcebase is None:
             mod.sourceHref = None
         else:
             projBaseDir = mod.system.options.projectbasedirectory
-            mod.sourceHref = (
-                self.sourcebase +
-                mod.filepath[len(projBaseDir):])
+            relative = source_path.relative_to(projBaseDir).as_posix()
+            mod.sourceHref = f'{self.sourcebase}/{relative}'
 
-    def addModule(self, modpath, modname, parentPackage=None):
-        mod = self.Module(self, modname, parentPackage)
+    def addModule(self,
+            modpath: Path,
+            modname: str,
+            parentPackage: Optional[_PackageT] = None
+            ) -> None:
+        mod = self.Module(self, modname, parentPackage, modpath)
         self.addObject(mod)
         self.progress(
             "addModule", len(self.allobjects),
             None, "modules and packages discovered")
-        mod.filepath = modpath
         self.unprocessed_modules.add(mod)
         self.module_count += 1
-        self.setSourceHref(mod)
+        self.setSourceHref(mod, modpath)
 
-    def ensureModule(self, module_full_name):
-        if module_full_name in self.allobjects:
-            return self.allobjects[module_full_name]
+    def ensureModule(self, module_full_name: str, modpath: Path) -> _ModuleT:
+        try:
+            module = self.allobjects[module_full_name]
+            assert isinstance(module, Module)
+        except KeyError:
+            pass
+        else:
+            return module
+
+        parent_package: Optional[_PackageT]
         if '.' in module_full_name:
             parent_name, module_name = module_full_name.rsplit('.', 1)
             parent_package = self.ensurePackage(parent_name)
         else:
             parent_package = None
             module_name = module_full_name
-        module = self.Module(self, module_name, parent_package)
+        module = self.Module(self, module_name, parent_package, modpath)
         self.addObject(module)
         return module
 
-    def ensurePackage(self, package_full_name):
+    def ensurePackage(self, package_full_name: str) -> _PackageT:
         if package_full_name in self.allobjects:
-            return self.allobjects[package_full_name]
+            package = self.allobjects[package_full_name]
+            assert isinstance(package, Package)
+            return package
+        parent_package: Optional[_PackageT]
         if '.' in package_full_name:
             parent_name, package_name = package_full_name.rsplit('.', 1)
             parent_package = self.ensurePackage(parent_name)
@@ -576,15 +746,17 @@ class System(object):
         self.addObject(package)
         return package
 
-    def _introspectThing(self, thing, parent, parentMod):
+    def _introspectThing(self, thing: object, parent: Documentable, parentMod: _ModuleT) -> None:
         for k, v in thing.__dict__.items():
-            if isinstance(v, (types.BuiltinFunctionType,
-                              types.FunctionType)):
+            if (isinstance(v, (types.BuiltinFunctionType, types.FunctionType))
+                    # In PyPy 7.3.1, functions from extensions are not
+                    # instances of the above abstract types.
+                    or v.__class__.__name__ == 'builtin_function_or_method'):
                 f = self.Function(self, k, parent)
                 f.parentMod = parentMod
                 f.docstring = v.__doc__
                 f.decorators = None
-                f.argspec = ((), None, None, ())
+                f.signature = Signature()
                 self.addObject(f)
             elif isinstance(v, type):
                 c = self.Class(self, k, parent)
@@ -596,15 +768,19 @@ class System(object):
                 self.addObject(c)
                 self._introspectThing(v, c, parentMod)
 
-    def introspectModule(self, py_mod, module_full_name):
-        module = self.ensureModule(module_full_name)
+    def introspectModule(self, path: Path, module_full_name: str) -> None:
+        spec = importlib.util.spec_from_file_location(module_full_name, path)
+        py_mod = importlib.util.module_from_spec(spec)
+        loader = spec.loader
+        assert isinstance(loader, importlib.abc.Loader), loader
+        loader.exec_module(py_mod)
+        module = self.ensureModule(module_full_name, path)
         module.docstring = py_mod.__doc__
         self._introspectThing(py_mod, module, module)
 
-    def addPackage(self, dirpath, parentPackage=None):
+    def addPackage(self, dirpath: str, parentPackage: Optional[_PackageT] = None) -> None:
         if not os.path.exists(dirpath):
-            raise Exception("package path %r does not exist!"
-                            %(dirpath,))
+            raise Exception(f"package path {dirpath!r} does not exist!")
         if not os.path.exists(os.path.join(dirpath, '__init__.py')):
             raise Exception("you must pass a package directory to "
                             "addPackage")
@@ -615,8 +791,6 @@ class System(object):
         package_name = os.path.basename(dirpath)
         package_full_name = prefix + package_name
         package = self.ensurePackage(package_full_name)
-        package.filepath = dirpath
-        self.setSourceHref(package)
         for fname in sorted(os.listdir(dirpath)):
             fullname = os.path.join(dirpath, fname)
             if os.path.isdir(fullname):
@@ -626,38 +800,34 @@ class System(object):
             elif not fname.startswith('.'):
                 self.addModuleFromPath(package, fullname)
 
-    def addModuleFromPath(self, package, path):
-        for (suffix, mode, type) in imp.get_suffixes():
+    def addModuleFromPath(self, package: Optional[_PackageT], path: str) -> None:
+        for suffix in importlib.machinery.all_suffixes():
             if not path.endswith(suffix):
                 continue
             module_name = os.path.basename(path[:-len(suffix)])
-            if type == imp.C_EXTENSION:
+            if suffix in importlib.machinery.EXTENSION_SUFFIXES:
                 if not self.options.introspect_c_modules:
                     continue
                 if package is not None:
-                    module_full_name = "%s.%s" % (
-                        package.fullName(), module_name)
+                    module_full_name = f'{package.fullName()}.{module_name}'
                 else:
                     module_full_name = module_name
-                py_mod = imp.load_module(
-                    module_full_name, open(path, 'rb'), path,
-                    (suffix, mode, type))
-                self.introspectModule(py_mod, module_full_name)
-            elif type == imp.PY_SOURCE:
-                self.addModule(path, module_name, package)
+                self.introspectModule(Path(path), module_full_name)
+            elif suffix in importlib.machinery.SOURCE_SUFFIXES:
+                self.addModule(Path(path), module_name, package)
             break
 
-    def handleDuplicate(self, obj):
+    def handleDuplicate(self, obj: Documentable) -> None:
         '''This is called when we see two objects with the same
-        .fullName(), for example:
+        .fullName(), for example::
 
-        class C:
-            if something:
-                def meth(self):
-                    implementation 1
-            else:
-                def meth(self):
-                    implementation 2
+            class C:
+                if something:
+                    def meth(self):
+                        implementation 1
+                else:
+                    def meth(self):
+                        implementation 2
 
         The default is that the second definition "wins".
         '''
@@ -666,15 +836,15 @@ class System(object):
         while (fullName + ' ' + str(i)) in self.allobjects:
             i += 1
         prev = self.allobjects[fullName]
-        self._warning(obj.parent, "duplicate", prev)
-        def remove(o):
+        self._warning(obj.parent, "duplicate", str(prev))
+        def remove(o: Documentable) -> None:
             del self.allobjects[o.fullName()]
             oc = list(o.contents.values())
             for c in oc:
                 remove(c)
         remove(prev)
         prev.name = obj.name + ' ' + str(i)
-        def readd(o):
+        def readd(o: Documentable) -> None:
             self.allobjects[o.fullName()] = o
             for c in o.contents.values():
                 readd(c)
@@ -682,28 +852,29 @@ class System(object):
         self.allobjects[fullName] = obj
 
 
-    def getProcessedModule(self, modname):
+    def getProcessedModule(self, modname: str) -> Optional[_ModuleT]:
         mod = self.allobjects.get(modname)
         if mod is None:
             return None
         if isinstance(mod, Package):
-            return self.getProcessedModule(modname + '.__init__').parent
+            return self.getProcessedModule(modname + '.__init__')
         if not isinstance(mod, Module):
             return None
 
         if mod.state is ProcessingState.UNPROCESSED:
             self.processModule(mod)
 
+        assert mod.state in (ProcessingState.PROCESSING, ProcessingState.PROCESSED)
         return mod
 
 
-    def processModule(self, mod):
+    def processModule(self, mod: _ModuleT) -> None:
         assert mod.state is ProcessingState.UNPROCESSED
         mod.state = ProcessingState.PROCESSING
-        if getattr(mod, 'filepath', None) is None:
+        if mod.source_path is None:
             return
         builder = self.defaultBuilder(self)
-        ast = builder.parseFile(mod.filepath)
+        ast = builder.parseFile(mod.source_path)
         if ast:
             self.processing_modules.append(mod.fullName())
             self.msg("processModule", "processing %s"%(self.processing_modules), 1)
@@ -716,17 +887,16 @@ class System(object):
             'process',
             self.module_count - len(self.unprocessed_modules),
             self.module_count,
-            "modules processed %s warnings"%(
-            sum(len(v) for v in self.warnings.values()),))
+            f"modules processed, {self.violations} warnings")
 
 
-    def process(self):
+    def process(self) -> None:
         while self.unprocessed_modules:
             mod = next(iter(self.unprocessed_modules))
             self.processModule(mod)
 
 
-    def fetchIntersphinxInventories(self, cache):
+    def fetchIntersphinxInventories(self, cache: CacheT) -> None:
         """
         Download and parse intersphinx inventories based on configuration.
         """
