@@ -1,12 +1,7 @@
-from __future__ import print_function
+from twisted.web.template import Element, XMLFile, renderer, tags
 
-from compiler import ast
-
-from twisted.web.template import Element, renderer, tags, XMLFile
-
-from pydoctor import ast_pp
 from pydoctor.templatewriter import util
-from pydoctor.templatewriter.pages import signature
+from pydoctor.templatewriter.pages import format_decorators, signature
 
 
 class FunctionChild(Element):
@@ -35,36 +30,15 @@ class FunctionChild(Element):
 
     @renderer
     def decorator(self, request, tag):
-
-        decorators = []
-
-        if self.ob.decorators:
-            for dec in self.ob.decorators:
-                if isinstance(dec.asList()[0], ast.Name):
-                    fn = self.ob.expandName(dec.asList()[0].name)
-                    # We don't want to show the deprecated decorator, it shows up
-                    # as an infobox
-                    if fn == "twisted.python.deprecate.deprecated":
-                        break
-                decorators.append(ast_pp.pp(dec))
-
-        if self.ob.kind == "Class Method" \
-               and 'classmethod' not in decorators:
-            decorators.append('classmethod')
-        elif self.ob.kind == "Static Method" \
-                 and 'staticmethod' not in decorators:
-            decorators.append('staticmethod')
-
-        if decorators:
-            decorator = [('@' + dec, tags.br()) for dec in decorators]
-        else:
-            decorator = ()
-
-        return decorator
+        return list(format_decorators(self.ob))
 
     @renderer
-    def functionName(self, request, tag):
-        return [self.ob.name, '(', signature(self.ob.argspec), '):']
+    def functionDef(self, request, tag):
+        def_stmt = 'async def' if self.ob.is_async else 'def'
+        name = self.ob.name
+        if name.endswith('.setter') or name.endswith('.deleter'):
+            name = name[:name.rindex('.')]
+        return [def_stmt, ' ', name, signature(self.ob), ':']
 
     @renderer
     def sourceLink(self, request, tag):
