@@ -1,9 +1,9 @@
-from __future__ import print_function
-
-from twisted.web.template import Element, renderer, TagLoader, XMLFile
+from twisted.web.template import Element, TagLoader, XMLFile, renderer
 
+from pydoctor.model import Function
 from pydoctor.templatewriter import util
 
+
 class TableRow(Element):
 
     def __init__(self, loader, docgetter, ob, child):
@@ -21,7 +21,13 @@ class TableRow(Element):
 
     @renderer
     def kind(self, request, tag):
-        return tag.clear()(self.child.kind)
+        child = self.child
+        kind = child.kind
+        if isinstance(child, Function) and child.is_async:
+            # The official name is "coroutine function", but that is both
+            # a bit long and not as widely recognized.
+            kind = f'Async {kind}'
+        return tag.clear()(kind)
 
     @renderer
     def name(self, request, tag):
