@@ -1,15 +1,16 @@
 """Badly named module that contains the driving code for the rendering."""
 
-from __future__ import print_function
+from abc import ABC
+from typing import Type
+import os
+import shutil
 
-from pydoctor.templatewriter.util import link, templatefile
-from pydoctor.templatewriter import DOCTYPE, pages, summary
 from pydoctor import model
-
+from pydoctor.templatewriter import DOCTYPE, pages, summary
+from pydoctor.templatewriter.util import templatefile
+from twisted.python.filepath import FilePath
 from twisted.web.template import flattenString
 
-import os, shutil
-
 
 def flattenToFile(fobj, page):
     fobj.write(DOCTYPE)
@@ -21,7 +22,15 @@ def flattenToFile(fobj, page):
         raise err[0]
 
 
-class TemplateWriter:
+class TemplateWriter(ABC):
+    @classmethod
+    def __subclasshook__(cls, subclass: Type[object]) -> bool:
+        for name in dir(cls):
+            if not name.startswith('_'):
+                if not hasattr(subclass, name):
+                    return False
+        return True
+
     def __init__(self, filebase):
         self.base = filebase
         self.written_pages = 0
@@ -29,8 +38,7 @@ class TemplateWriter:
         self.dry_run = False
 
     def prepOutputDirectory(self):
-        if not os.path.exists(self.base):
-            os.mkdir(self.base)
+        os.makedirs(self.base, exist_ok=True)
         shutil.copyfile(templatefile('apidocs.css'),
                         os.path.join(self.base, 'apidocs.css'))
         shutil.copyfile(templatefile('bootstrap.min.css'),
@@ -41,10 +49,10 @@ class TemplateWriter:
     def writeIndividualFiles(self, obs, functionpages=False):
         self.dry_run = True
         for ob in obs:
-            self.writeDocsFor(ob, functionpages=functionpages)
+            self._writeDocsFor(ob, functionpages=functionpages)
         self.dry_run = False
         for ob in obs:
-            self.writeDocsFor(ob, functionpages=functionpages)
+            self._writeDocsFor(ob, functionpages=functionpages)
 
     def writeModuleIndex(self, system):
         import time
@@ -52,26 +60,25 @@ class TemplateWriter:
             system.msg('html', 'starting ' + pclass.__name__ + ' ...', nonl=True)
             T = time.time()
             page = pclass(system)
-            f = open(os.path.join(self.base, pclass.filename), 'w')
+            f = open(os.path.join(self.base, pclass.filename), 'wb')
             flattenToFile(f, page)
             f.close()
             system.msg('html', "took %fs"%(time.time() - T), wantsnl=False)
 
-    def writeDocsFor(self, ob, functionpages):
+    def _writeDocsFor(self, ob, functionpages):
         if not ob.isVisible:
             return
-        isfunc = ob.documentation_location == model.DocLocation.PARENT_PAGE
-        if (isfunc and functionpages) or not isfunc:
+        if functionpages or ob.documentation_location is model.DocLocation.OWN_PAGE:
             if self.dry_run:
                 self.total_pages += 1
             else:
-                f = open(os.path.join(self.base, link(ob)), 'w')
-                self.writeDocsForOne(ob, f)
-                f.close()
-        for o in ob.orderedcontents:
-            self.writeDocsFor(o, functionpages)
+                path = FilePath(self.base).child(f'{ob.fullName()}.html')
+                with path.open('wb') as out:
+                    self._writeDocsForOne(ob, out)
+        for o in ob.contents.values():
+            self._writeDocsFor(o, functionpages)
 
-    def writeDocsForOne(self, ob, fobj):
+    def _writeDocsForOne(self, ob, fobj):
         if not ob.isVisible:
             return
         # brrrrrrr!
@@ -83,8 +90,8 @@ class TemplateWriter:
                 break
         else:
             pclass = pages.CommonPage
-        self.system.msg('html', str(ob), thresh=1)
+        ob.system.msg('html', str(ob), thresh=1)
         page = pclass(ob)
         self.written_pages += 1
-        self.system.progress('html', self.written_pages, self.total_pages, 'pages written')
+        ob.system.progress('html', self.written_pages, self.total_pages, 'pages written')
         flattenToFile(fobj, page)
