@@ -1,5 +1,5 @@
-from __future__ import print_function
-
+from typing import Optional, Tuple, Type, overload
+import ast
 import textwrap
 
 import astor
@@ -7,11 +7,12 @@ import astor
 from twisted.python._pydoctor import TwistedSystem
 
 from pydoctor import astbuilder, model
-from pydoctor.epydoc.markup import flatten
-from pydoctor.epydoc2stan import get_parsed_type
+from pydoctor.epydoc.markup import DocstringLinker, ParsedDocstring, flatten
+from pydoctor.epydoc.markup.epytext import ParsedEpytextDocstring
+from pydoctor.epydoc2stan import format_summary, get_parsed_type
 from pydoctor.zopeinterface import ZopeInterfaceSystem
 
-from . import py2only, py3only, typecomment
+from . import CapSys, posonlyargs, typecomment
 import pytest
 
 
@@ -19,9 +20,14 @@ systemcls_param = pytest.mark.parametrize(
     'systemcls', (model.System, ZopeInterfaceSystem, TwistedSystem)
     )
 
-def fromText(text, modname='<test>', system=None,
-             buildercls=None,
-             systemcls=model.System):
+def fromAST(
+        ast: ast.Module,
+        modname: str = '<test>',
+        parent_name: Optional[str] = None,
+        system: Optional[model.System] = None,
+        buildercls: Optional[Type[astbuilder.ASTBuilder]] = None,
+        systemcls: Type[model.System] = model.System
+        ) -> model.Module:
     if system is None:
         _system = systemcls()
     else:
@@ -29,35 +35,80 @@ def fromText(text, modname='<test>', system=None,
     if buildercls is None:
         buildercls = _system.defaultBuilder
     builder = buildercls(_system)
-    mod = builder._push(_system.Module, modname, None)
+    if parent_name is None:
+        full_name = modname
+    else:
+        full_name = f'{parent_name}.{modname}'
+        # Set containing package as parent.
+        builder.current = _system.allobjects[parent_name]
+    mod: model.Module = builder._push(_system.Module, modname, None)
     builder._pop(_system.Module)
-    ast = astbuilder.parse(textwrap.dedent(text))
     builder.processModuleAST(ast, mod)
-    mod = _system.allobjects[modname]
-    mod.ast = ast
+    assert mod is _system.allobjects[full_name]
     mod.state = model.ProcessingState.PROCESSED
     return mod
 
-def unwrap(parsed_docstring):
+def fromText(
+        text: str,
+        *,
+        modname: str = '<test>',
+        parent_name: Optional[str] = None,
+        system: Optional[model.System] = None,
+        buildercls: Optional[Type[astbuilder.ASTBuilder]] = None,
+        systemcls: Type[model.System] = model.System
+        ) -> model.Module:
+    ast = astbuilder._parse(textwrap.dedent(text))
+    return fromAST(ast, modname, parent_name, system, buildercls, systemcls)
+
+def unwrap(parsed_docstring: ParsedEpytextDocstring) -> str:
     epytext = parsed_docstring._tree
+    assert epytext is not None
     assert epytext.tag == 'epytext'
     assert len(epytext.children) == 1
     para = epytext.children[0]
     assert para.tag == 'para'
     assert len(para.children) == 1
-    return para.children[0]
+    value = para.children[0]
+    assert isinstance(value, str)
+    return value
 
-def to_html(parsed_docstring):
-    return flatten(parsed_docstring.to_stan(None))
+class NotFoundLinker(DocstringLinker):
+    """A DocstringLinker implementation that cannot find any links."""
 
-def type2str(type_expr):
+    def resolve_identifier(self, identifier: str) -> Optional[str]:
+        return None
+
+    def resolve_identifier_xref(self, identifier: str, lineno: int) -> str:
+        raise LookupError(identifier)
+
+def to_html(parsed_docstring: ParsedDocstring) -> str:
+    return flatten(parsed_docstring.to_stan(NotFoundLinker()))
+
+@overload
+def type2str(type_expr: None) -> None: ...
+
+@overload
+def type2str(type_expr: ast.expr) -> str: ...
+
+def type2str(type_expr: Optional[ast.expr]) -> Optional[str]:
     if type_expr is None:
         return None
     else:
-        return astor.to_source(type_expr).strip()
+        src = astor.to_source(type_expr)
+        assert isinstance(src, str)
+        return src.strip()
+
+def ann_str_and_line(obj: model.Documentable) -> Tuple[str, int]:
+    """Return the textual representation and line number of an object's
+    type annotation.
+    @param obj: Documentable object with a type annotation.
+    """
+    ann = obj.annotation # type: ignore[attr-defined]
+    assert ann is not None
+    return type2str(ann), ann.lineno
 
 @systemcls_param
-def test_no_docstring(systemcls):
+def test_no_docstring(systemcls: Type[model.System]) -> None:
     # Inheritance of the docstring of an overridden method depends on
     # methods with no docstring having None in their 'docstring' field.
     mod = fromText('''
@@ -73,7 +124,7 @@ def test_no_docstring(systemcls):
     assert m.docstring is None
 
 @systemcls_param
-def test_simple(systemcls):
+def test_function_simple(systemcls: Type[model.System]) -> None:
     src = '''
     """ MOD DOC """
     def f():
@@ -84,32 +135,81 @@ def test_simple(systemcls):
     func, = mod.contents.values()
     assert func.fullName() == '<test>.f'
     assert func.docstring == """This is a docstring."""
+    assert func.is_async is False
 
 
 @systemcls_param
-def test_function_argspec(systemcls):
-    src = textwrap.dedent('''
-    def f(a, b=3, *c, **kw):
-        pass
-    ''')
+def test_function_async(systemcls: Type[model.System]) -> None:
+    src = '''
+    """ MOD DOC """
+    async def a():
+        """This is a docstring."""
+    '''
     mod = fromText(src, systemcls=systemcls)
-    docfunc, = mod.contents.values()
-    assert docfunc.argspec == (['a', 'b'], 'c', 'kw', ('3',))
+    assert len(mod.contents) == 1
+    func, = mod.contents.values()
+    assert func.fullName() == '<test>.a'
+    assert func.docstring == """This is a docstring."""
+    assert func.is_async is True
 
 
-@py2only
+@pytest.mark.parametrize('signature', (
+    '()',
+    '(*, a, b=None)',
+    '(*, a=(), b)',
+    '(a, b=3, *c, **kw)',
+    '(f=True)',
+    '(x=0.1, y=-2)',
+    '(s=\'theory\', t="con\'text")',
+    ))
 @systemcls_param
-def test_function_argspec_with_tuple(systemcls):
-    src = textwrap.dedent('''
-    def f((a,z), b=3, *c, **kw):
-        pass
-    ''')
-    mod = fromText(src, systemcls=systemcls)
+def test_function_signature(signature: str, systemcls: Type[model.System]) -> None:
+    """A round trip from source to inspect.Signature and back produces
+    the original text.
+    """
+    mod = fromText(f'def f{signature}: ...', systemcls=systemcls)
     docfunc, = mod.contents.values()
-    assert docfunc.argspec ==  ([['a', 'z'], 'b'], 'c', 'kw', ('3',))
+    assert isinstance(docfunc, model.Function)
+    assert str(docfunc.signature) == signature
+
+@posonlyargs
+@pytest.mark.parametrize('signature', (
+    '(x, y, /)',
+    '(x, y=0, /)',
+    '(x, y, /, z, w)',
+    '(x, y, /, z, w=42)',
+    '(x, y, /, z=0, w=0)',
+    '(x, y=3, /, z=5, w=7)',
+    '(x, /, *v, a=1, b=2)',
+    '(x, /, *, a=1, b=2, **kwargs)',
+    ))
+@systemcls_param
+def test_function_signature_posonly(signature: str, systemcls: Type[model.System]) -> None:
+    test_function_signature(signature, systemcls)
+
+
+@pytest.mark.parametrize('signature', (
+    '(a, a)',
+    ))
+@systemcls_param
+def test_function_badsig(signature: str, systemcls: Type[model.System], capsys: CapSys) -> None:
+    """When a function has an invalid signature, an error is logged and
+    the empty signature is returned.
+
+    Note that most bad signatures lead to a SyntaxError, which we cannot
+    recover from. This test checks what happens if the AST can be produced
+    but inspect.Signature() rejects the parsed parameters.
+    """
+    mod = fromText(f'def f{signature}: ...', systemcls=systemcls, modname='mod')
+    docfunc, = mod.contents.values()
+    assert isinstance(docfunc, model.Function)
+    assert str(docfunc.signature) == '()'
+    captured = capsys.readouterr().out
+    assert captured.startswith("mod:1: mod.f has invalid parameters: ")
+
 
 @systemcls_param
-def test_class(systemcls):
+def test_class(systemcls: Type[model.System]) -> None:
     src = '''
     class C:
         def f():
@@ -127,7 +227,7 @@ def test_class(systemcls):
 
 
 @systemcls_param
-def test_class_with_base(systemcls):
+def test_class_with_base(systemcls: Type[model.System]) -> None:
     src = '''
     class C:
         def f():
@@ -152,7 +252,7 @@ def test_class_with_base(systemcls):
     assert base == '<test>.C'
 
 @systemcls_param
-def test_follow_renaming(systemcls):
+def test_follow_renaming(systemcls: Type[model.System]) -> None:
     src = '''
     class C: pass
     D = C
@@ -164,7 +264,7 @@ def test_follow_renaming(systemcls):
     assert E.baseobjects == [C], E.baseobjects
 
 @systemcls_param
-def test_relative_import_past_root(systemcls, capsys):
+def test_relative_import_past_root(systemcls: Type[model.System], capsys: CapSys) -> None:
     src = '''
     from ..X import A
     '''
@@ -173,7 +273,7 @@ def test_relative_import_past_root(systemcls, capsys):
     assert "relative import level too high" in captured
 
 @systemcls_param
-def test_class_with_base_from_module(systemcls):
+def test_class_with_base_from_module(systemcls: Type[model.System]) -> None:
     src = '''
     from X.Y import A
     from Z import B as C
@@ -216,31 +316,42 @@ def test_class_with_base_from_module(systemcls):
     assert base3 == 'Y.Z.C', base3
 
 @systemcls_param
-def test_aliasing(systemcls):
-    def addsrc(system):
-        src_a = '''
+def test_aliasing(systemcls: Type[model.System]) -> None:
+    def addsrc(system: model.System) -> None:
+        src_private = '''
         class A:
             pass
         '''
-        src_b = '''
-        from a import A as B
+        src_export = '''
+        from _private import A as B
+        __all__ = ['B']
         '''
-        src_c = '''
-        from b import B
+        src_user = '''
+        from public import B
         class C(B):
             pass
         '''
-        fromText(src_a, 'a', system)
-        fromText(src_b, 'b', system)
-        fromText(src_c, 'c', system)
+        fromText(src_private, modname='_private', system=system)
+        fromText(src_export, modname='public', system=system)
+        fromText(src_user, modname='app', system=system)
 
     system = systemcls()
     addsrc(system)
-    assert system.allobjects['c.C'].bases == ['a.A']
+    C = system.allobjects['app.C']
+    assert isinstance(C, model.Class)
+    # An older version of this test expected _private.A as the result.
+    # The expected behavior was changed because:
+    # - relying on on-demand processing of other modules is unreliable when
+    #   there are cyclic imports: expandName() on a module that is still being
+    #   processed can return the not-found result for a name that does exist
+    # - code should be importing names from their official home, so if we
+    #   import public.B then for the purposes of documentation public.B is
+    #   the name we should use
+    assert C.bases == ['public.B']
 
 @systemcls_param
-def test_more_aliasing(systemcls):
-    def addsrc(system):
+def test_more_aliasing(systemcls: Type[model.System]) -> None:
+    def addsrc(system: model.System) -> None:
         src_a = '''
         class A:
             pass
@@ -256,17 +367,21 @@ def test_more_aliasing(systemcls):
         class D(C):
             pass
         '''
-        fromText(src_a, 'a', system)
-        fromText(src_b, 'b', system)
-        fromText(src_c, 'c', system)
-        fromText(src_d, 'd', system)
+        fromText(src_a, modname='a', system=system)
+        fromText(src_b, modname='b', system=system)
+        fromText(src_c, modname='c', system=system)
+        fromText(src_d, modname='d', system=system)
 
     system = systemcls()
     addsrc(system)
-    assert system.allobjects['d.D'].bases == ['a.A']
+    D = system.allobjects['d.D']
+    assert isinstance(D, model.Class)
+    # An older version of this test expected a.A as the result.
+    # Read the comment in test_aliasing() to learn why this was changed.
+    assert D.bases == ['c.C']
 
 @systemcls_param
-def test_aliasing_recursion(systemcls):
+def test_aliasing_recursion(systemcls: Type[model.System]) -> None:
     system = systemcls()
     src = '''
     class C:
@@ -275,11 +390,11 @@ def test_aliasing_recursion(systemcls):
     class D(C):
         pass
     '''
-    mod = fromText(src, 'mod', system)
+    mod = fromText(src, modname='mod', system=system)
     assert mod.contents['D'].bases == ['mod.C'], mod.contents['D'].bases
 
 @systemcls_param
-def test_documented_no_alias(systemcls):
+def test_documented_no_alias(systemcls: Type[model.System]) -> None:
     """A variable that is documented should not be considered an alias."""
     # TODO: We should also verify this for inline docstrings, but the code
     #       currently doesn't support that. We should perhaps store aliases
@@ -302,7 +417,7 @@ def test_documented_no_alias(systemcls):
     assert f.linenumber
 
 @systemcls_param
-def test_subclasses(systemcls):
+def test_subclasses(systemcls: Type[model.System]) -> None:
     src = '''
     class A:
         pass
@@ -310,11 +425,12 @@ def test_subclasses(systemcls):
         pass
     '''
     system = fromText(src, systemcls=systemcls).system
-    assert (system.allobjects['<test>.A'].subclasses ==
-            [system.allobjects['<test>.B']])
+    A = system.allobjects['<test>.A']
+    assert isinstance(A, model.Class)
+    assert A.subclasses == [system.allobjects['<test>.B']]
 
 @systemcls_param
-def test_inherit_names(systemcls):
+def test_inherit_names(systemcls: Type[model.System]) -> None:
     src = '''
     class A:
         pass
@@ -325,7 +441,7 @@ def test_inherit_names(systemcls):
     assert [b.name for b in mod.contents['A'].allbases()] == ['A 0']
 
 @systemcls_param
-def test_nested_class_inheriting_from_same_module(systemcls):
+def test_nested_class_inheriting_from_same_module(systemcls: Type[model.System]) -> None:
     src = '''
     class A:
         pass
@@ -336,27 +452,29 @@ def test_nested_class_inheriting_from_same_module(systemcls):
     fromText(src, systemcls=systemcls)
 
 @systemcls_param
-def test_all_recognition(systemcls):
-    mod = fromText('''
+def test_all_recognition(systemcls: Type[model.System]) -> None:
+    ast = astbuilder._parse(textwrap.dedent('''
     def f():
         pass
     __all__ = ['f']
-    ''', systemcls=systemcls)
-    astbuilder.findAll(mod.ast, mod)
+    '''))
+    mod = fromAST(ast, systemcls=systemcls)
+    astbuilder.findAll(ast, mod)
     assert mod.all == ['f']
     assert '__all__' not in mod.contents
 
 @systemcls_param
-def test_all_in_class_non_recognition(systemcls):
-    mod = fromText('''
+def test_all_in_class_non_recognition(systemcls: Type[model.System]) -> None:
+    ast = astbuilder._parse(textwrap.dedent('''
     class C:
         __all__ = ['f']
-    ''', systemcls=systemcls)
-    astbuilder.findAll(mod.ast, mod)
+    '''))
+    mod = fromAST(ast, systemcls=systemcls)
+    astbuilder.findAll(ast, mod)
     assert mod.all is None
 
 @systemcls_param
-def test_classmethod(systemcls):
+def test_classmethod(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         @classmethod
@@ -373,7 +491,7 @@ def test_classmethod(systemcls):
     assert mod.contents['C'].contents['f'].kind == 'Class Method'
 
 @systemcls_param
-def test_classdecorator(systemcls):
+def test_classdecorator(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     def cd(cls):
         pass
@@ -387,7 +505,7 @@ def test_classdecorator(systemcls):
 
 
 @systemcls_param
-def test_classdecorator_with_args(systemcls):
+def test_classdecorator_with_args(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     def cd(): pass
     class A: pass
@@ -403,7 +521,7 @@ def test_classdecorator_with_args(systemcls):
 
 
 @systemcls_param
-def test_methoddecorator(systemcls, capsys):
+def test_methoddecorator(systemcls: Type[model.System], capsys: CapSys) -> None:
     mod = fromText('''
     class C:
         def method_undecorated():
@@ -431,7 +549,68 @@ def test_methoddecorator(systemcls, capsys):
 
 
 @systemcls_param
-def test_import_star(systemcls):
+def test_assignment_to_method_in_class(systemcls: Type[model.System]) -> None:
+    """An assignment to a method in a class body does not change the type
+    of the documentable.
+
+    If the name we assign to exists and it does not belong to an Attribute
+    (it's a Function instead, in this test case), the assignment will be
+    ignored.
+    """
+    mod = fromText('''
+    class Base:
+        def base_method():
+            """Base method docstring."""
+
+    class Sub(Base):
+        base_method = wrap_method(base_method)
+        """Overriding the docstring is not supported."""
+
+        def sub_method():
+            """Sub method docstring."""
+        sub_method = wrap_method(sub_method)
+        """Overriding the docstring is not supported."""
+    ''', systemcls=systemcls)
+    assert isinstance(mod.contents['Base'].contents['base_method'], model.Function)
+    assert mod.contents['Sub'].contents.get('base_method') is None
+    sub_method = mod.contents['Sub'].contents['sub_method']
+    assert isinstance(sub_method, model.Function)
+    assert sub_method.docstring == """Sub method docstring."""
+
+
+@systemcls_param
+def test_assignment_to_method_in_init(systemcls: Type[model.System]) -> None:
+    """An assignment to a method inside __init__() does not change the type
+    of the documentable.
+
+    If the name we assign to exists and it does not belong to an Attribute
+    (it's a Function instead, in this test case), the assignment will be
+    ignored.
+    """
+    mod = fromText('''
+    class Base:
+        def base_method():
+            """Base method docstring."""
+
+    class Sub(Base):
+        def sub_method():
+            """Sub method docstring."""
+
+        def __init__(self):
+            self.base_method = wrap_method(self.base_method)
+            """Overriding the docstring is not supported."""
+            self.sub_method = wrap_method(self.sub_method)
+            """Overriding the docstring is not supported."""
+    ''', systemcls=systemcls)
+    assert isinstance(mod.contents['Base'].contents['base_method'], model.Function)
+    assert mod.contents['Sub'].contents.get('base_method') is None
+    sub_method = mod.contents['Sub'].contents['sub_method']
+    assert isinstance(sub_method, model.Function)
+    assert sub_method.docstring == """Sub method docstring."""
+
+
+@systemcls_param
+def test_import_star(systemcls: Type[model.System]) -> None:
     mod_a = fromText('''
     def f(): pass
     ''', modname='a', systemcls=systemcls)
@@ -442,7 +621,66 @@ def test_import_star(systemcls):
 
 
 @systemcls_param
-def test_inline_docstring_modulevar(systemcls):
+def test_import_func_from_package(systemcls: Type[model.System]) -> None:
+    """Importing a function from a package should look in the C{__init__}
+    module.
+
+    In this test the following hierarchy is constructed::
+
+        package a
+          module __init__
+            defines function 'f'
+        module b
+          imports function 'f'
+
+    We verify that when module C{b} imports the name C{f} from package C{a},
+    it imports the function C{f} from the module C{a.__init__}.
+    """
+    system = systemcls()
+    system.ensurePackage('a')
+    mod_a = fromText('''
+    def f(): pass
+    ''', modname='__init__', parent_name='a', system=system)
+    mod_b = fromText('''
+    from a import f
+    ''', modname='b', system=system)
+    assert mod_b.resolveName('f') == mod_a.contents['f']
+
+
+@systemcls_param
+def test_import_module_from_package(systemcls: Type[model.System]) -> None:
+    """Importing a module from a package should not look in C{__init__}
+    module.
+
+    In this test the following hierarchy is constructed::
+
+        package a
+          module __init__
+          module b
+            defines function 'f'
+        module c
+          imports module 'a.b'
+
+    We verify that when module C{c} imports the name C{b} from package C{a},
+    it imports the module C{a.b} which contains C{f}.
+    """
+    system = systemcls()
+    system.ensurePackage('a')
+    fromText('''
+    # This module intentionally left blank.
+    ''', modname='__init__', parent_name='a', system=system)
+    mod_b = fromText('''
+    def f(): pass
+    ''', modname='b', parent_name='a', system=system)
+    mod_c = fromText('''
+    from a import b
+    f = b.f
+    ''', modname='c', system=system)
+    assert mod_c.resolveName('f') == mod_b.contents['f']
+
+
+@systemcls_param
+def test_inline_docstring_modulevar(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     """regular module docstring
 
@@ -469,7 +707,7 @@ def test_inline_docstring_modulevar(systemcls):
     assert not f.docstring
 
 @systemcls_param
-def test_inline_docstring_classvar(systemcls):
+def test_inline_docstring_classvar(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         """regular class docstring"""
@@ -500,9 +738,8 @@ def test_inline_docstring_classvar(systemcls):
     assert b.docstring == """inline doc for _b"""
     assert b.privacyClass is model.PrivacyClass.PRIVATE
 
-@py3only
 @systemcls_param
-def test_inline_docstring_annotated_classvar(systemcls):
+def test_inline_docstring_annotated_classvar(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         """regular class docstring"""
@@ -523,7 +760,7 @@ def test_inline_docstring_annotated_classvar(systemcls):
     assert b.privacyClass is model.PrivacyClass.PRIVATE
 
 @systemcls_param
-def test_inline_docstring_instancevar(systemcls):
+def test_inline_docstring_instancevar(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         """regular class docstring"""
@@ -584,9 +821,8 @@ def test_inline_docstring_instancevar(systemcls):
     assert f.privacyClass is model.PrivacyClass.VISIBLE
     assert f.kind == 'Instance Variable'
 
-@py3only
 @systemcls_param
-def test_inline_docstring_annotated_instancevar(systemcls):
+def test_inline_docstring_annotated_instancevar(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         """regular class docstring"""
@@ -608,7 +844,7 @@ def test_inline_docstring_annotated_instancevar(systemcls):
     assert b.docstring == """inline doc for b"""
 
 @systemcls_param
-def test_docstring_assignment(systemcls, capsys):
+def test_docstring_assignment(systemcls: Type[model.System], capsys: CapSys) -> None:
     mod = fromText('''
     def fun():
         pass
@@ -648,19 +884,19 @@ def test_docstring_assignment(systemcls, capsys):
     captured = capsys.readouterr()
     lines = captured.out.split('\n')
     assert len(lines) > 0 and lines[0] == \
-        "<unknown>:20: Unable to figure out target for __doc__ assignment"
+        "<test>:20: Unable to figure out target for __doc__ assignment"
     assert len(lines) > 1 and lines[1] == \
-        "<unknown>:21: Unable to figure out target for __doc__ assignment: " \
+        "<test>:21: Unable to figure out target for __doc__ assignment: " \
         "computed full name not found: real"
     assert len(lines) > 2 and lines[2] == \
-        "<unknown>:22: Unable to figure out value for __doc__ assignment, " \
+        "<test>:22: Unable to figure out value for __doc__ assignment, " \
         "maybe too complex"
     assert len(lines) > 3 and lines[3] == \
-        "<unknown>:23: Ignoring value assigned to __doc__: not a string"
+        "<test>:23: Ignoring value assigned to __doc__: not a string"
     assert len(lines) == 5 and lines[-1] == ''
 
 @systemcls_param
-def test_variable_scopes(systemcls):
+def test_variable_scopes(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     l = 1
     """module-level l"""
@@ -709,7 +945,7 @@ def test_variable_scopes(systemcls):
     assert m2.docstring == """class-level m"""
 
 @systemcls_param
-def test_variable_types(systemcls):
+def test_variable_types(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         """class docstring
@@ -785,9 +1021,8 @@ def test_variable_types(systemcls):
     assert str(unwrap(g.parsed_type)) == 'string'
     assert g.kind == 'Instance Variable'
 
-@py3only
 @systemcls_param
-def test_annotated_variables(systemcls):
+def test_annotated_variables(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         """class docstring
@@ -825,59 +1060,96 @@ def test_annotated_variables(systemcls):
     m: bytes = b"M"
     """module-level"""
     ''', modname='test', systemcls=systemcls)
+
+    def type2html(obj: model.Documentable) -> str:
+        parsed_type = get_parsed_type(obj)
+        assert parsed_type is not None
+        return to_html(parsed_type)
+
     C = mod.contents['C']
     a = C.contents['a']
     assert unwrap(a.parsed_docstring) == """first"""
-    assert to_html(get_parsed_type(a)) == 'string'
+    assert type2html(a) == 'string'
     b = C.contents['b']
     assert unwrap(b.parsed_docstring) == """second"""
-    assert to_html(get_parsed_type(b)) == 'string'
+    assert type2html(b) == 'string'
     c = C.contents['c']
     assert c.docstring == """third"""
-    assert to_html(get_parsed_type(c)) == '<code>str</code>'
+    assert type2html(c) == '<code>str</code>'
     d = C.contents['d']
     assert d.docstring == """fourth"""
-    assert to_html(get_parsed_type(d)) == '<code>str</code>'
+    assert type2html(d) == '<code>str</code>'
     e = C.contents['e']
     assert e.docstring == """fifth"""
-    assert to_html(get_parsed_type(e)) == '<code>List[C]</code>'
+    assert type2html(e) == '<code>List[C]</code>'
     f = C.contents['f']
     assert f.docstring == """sixth"""
-    assert to_html(get_parsed_type(f)) == '<code>List[C]</code>'
+    assert type2html(f) == '<code>List[C]</code>'
     g = C.contents['g']
     assert g.docstring == """seventh"""
-    assert to_html(get_parsed_type(g)) == '<code>List[C]</code>'
+    assert type2html(g) == '<code>List[C]</code>'
     s = C.contents['s']
     assert s.docstring == """instance"""
-    assert to_html(get_parsed_type(s)) == '<code>List[str]</code>'
+    assert type2html(s) == '<code>List[str]</code>'
     m = mod.contents['m']
     assert m.docstring == """module-level"""
-    assert to_html(get_parsed_type(m)) == '<code>bytes</code>'
+    assert type2html(m) == '<code>bytes</code>'
 
 @typecomment
 @systemcls_param
-def test_type_comment(systemcls, capsys):
+def test_type_comment(systemcls: Type[model.System], capsys: CapSys) -> None:
     mod = fromText('''
-    d = {} # type: Dict[str, int]
+    d = {} # type: dict[str, int]
     i = [] # type: ignore[misc]
     ''', systemcls=systemcls)
-    assert type2str(mod.contents['d'].annotation) == 'Dict[str, int]'
+    assert type2str(mod.contents['d'].annotation) == 'dict[str, int]'
     # We don't use ignore comments for anything at the moment,
     # but do verify that their presence doesn't break things.
-    assert type2str(mod.contents['i'].annotation) == 'List'
+    assert type2str(mod.contents['i'].annotation) == 'list'
     assert not capsys.readouterr().out
 
-@py3only
 @systemcls_param
-def test_bad_string_annotation(systemcls, capsys):
+def test_unstring_annotation(systemcls: Type[model.System]) -> None:
+    """Annotations or parts thereof that are strings are parsed and
+    line number information is preserved.
+    """
     mod = fromText('''
-    x: "["
+    a: "int"
+    b: 'str' = 'B'
+    c: list["Thingy"]
+    ''', systemcls=systemcls)
+    assert ann_str_and_line(mod.contents['a']) == ('int', 2)
+    assert ann_str_and_line(mod.contents['b']) == ('str', 3)
+    assert ann_str_and_line(mod.contents['c']) == ('list[Thingy]', 4)
+
+@pytest.mark.parametrize('annotation', ("[", "pass", "1 ; 2"))
+@systemcls_param
+def test_bad_string_annotation(
+        annotation: str, systemcls: Type[model.System], capsys: CapSys
+        ) -> None:
+    """Invalid string annotations must be reported as syntax errors."""
+    mod = fromText(f'''
+    x: "{annotation}"
     ''', modname='test', systemcls=systemcls)
-    assert mod.contents['x'].annotation is None
+    assert isinstance(mod.contents['x'].annotation, ast.expr)
     assert "syntax error in annotation" in capsys.readouterr().out
 
+@pytest.mark.parametrize('annotation,expected', (
+    ("Literal['[', ']']", "Literal['[', ']']"),
+    ("typing.Literal['pass', 'raise']", "typing.Literal['pass', 'raise']"),
+    ("Optional[Literal['1 ; 2']]", "Optional[Literal['1 ; 2']]"),
+    ("'Literal'['!']", "Literal['!']"),
+    (r"'Literal[\'if\', \'while\']'", "Literal['if', 'while']"),
+    ))
+def test_literal_string_annotation(annotation: str, expected: str) -> None:
+    """Strings inside Literal annotations must not be recursively parsed."""
+    stmt, = ast.parse(annotation).body
+    assert isinstance(stmt, ast.Expr)
+    unstringed = astbuilder._AnnotationStringParser().visit(stmt.value)
+    assert astor.to_source(unstringed).strip() == expected
+
 @systemcls_param
-def test_inferred_variable_types(systemcls):
+def test_inferred_variable_types(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     class C:
         a = "A"
@@ -899,19 +1171,17 @@ def test_inferred_variable_types(systemcls):
     m = b'octets'
     ''', modname='test', systemcls=systemcls)
     C = mod.contents['C']
-    assert type2str(C.contents['a'].annotation) == 'str'
-    assert type2str(C.contents['b'].annotation) == 'int'
-    assert type2str(C.contents['c'].annotation) == 'List[str]'
-    assert type2str(C.contents['d'].annotation) == 'Dict[str, int]'
-    assert type2str(C.contents['e'].annotation) == 'Tuple[bool, ...]'
-    assert type2str(C.contents['f'].annotation) == 'float'
-    # The Python 2.7 implementation of literal_eval() does not support
-    # set literals.
-    assert type2str(C.contents['g'].annotation) in ('Set[int]', None)
+    assert ann_str_and_line(C.contents['a']) == ('str', 3)
+    assert ann_str_and_line(C.contents['b']) == ('int', 4)
+    assert ann_str_and_line(C.contents['c']) == ('list[str]', 5)
+    assert ann_str_and_line(C.contents['d']) == ('dict[str, int]', 6)
+    assert ann_str_and_line(C.contents['e']) == ('tuple[bool, ...]', 7)
+    assert ann_str_and_line(C.contents['f']) == ('float', 8)
+    assert ann_str_and_line(C.contents['g']) == ('set[int]', 9)
     # Element type is unknown, not uniform or too complex.
-    assert type2str(C.contents['h'].annotation) == 'List'
-    assert type2str(C.contents['i'].annotation) == 'List'
-    assert type2str(C.contents['j'].annotation) == 'Tuple'
+    assert ann_str_and_line(C.contents['h']) == ('list', 10)
+    assert ann_str_and_line(C.contents['i']) == ('list', 11)
+    assert ann_str_and_line(C.contents['j']) == ('tuple', 12)
     # It is unlikely that a variable actually will contain only None,
     # so we should treat this as not be able to infer the type.
     assert C.contents['n'].annotation is None
@@ -921,15 +1191,13 @@ def test_inferred_variable_types(systemcls):
     assert C.contents['y'].annotation is None
     # Type inference isn't different for module and instance variables,
     # so we don't need to re-test everything.
-    assert type2str(C.contents['s'].annotation) == 'List[str]'
+    assert ann_str_and_line(C.contents['s']) == ('list[str]', 17)
     # Check that type is inferred on assignments with multiple targets.
-    assert type2str(C.contents['t'].annotation) == 'str'
-    # On Python 2.7, bytes literals are parsed into ast.Str objects,
-    # so there is no way to tell them apart from ASCII strings.
-    assert type2str(mod.contents['m'].annotation) in ('bytes', 'str')
+    assert ann_str_and_line(C.contents['t']) == ('str', 18)
+    assert ann_str_and_line(mod.contents['m']) == ('bytes', 19)
 
 @systemcls_param
-def test_type_from_attrib(systemcls):
+def test_type_from_attrib(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     import attr
     from attr import attrib
@@ -946,14 +1214,133 @@ def test_type_from_attrib(systemcls):
     assert type2str(C.contents['d'].annotation) == 'bool'
 
 @systemcls_param
-def test_detupling_assignment(systemcls):
+def test_detupling_assignment(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     a, b, c = range(3)
     ''', modname='test', systemcls=systemcls)
     assert sorted(mod.contents.keys()) == ['a', 'b', 'c']
 
 @systemcls_param
-def test_ignore_function_contents(systemcls):
+def test_property_decorator(systemcls: Type[model.System]) -> None:
+    """A function decorated with '@property' is documented as an attribute."""
+    mod = fromText('''
+    class C:
+        @property
+        def prop(self) -> str:
+            """For sale."""
+            return 'seaside'
+        @property
+        def oldschool(self):
+            """
+            @return: For rent.
+            @rtype: string
+            @see: U{https://example.com/}
+            """
+            return 'downtown'
+    ''', modname='test', systemcls=systemcls)
+    C = mod.contents['C']
+
+    prop = C.contents['prop']
+    assert isinstance(prop, model.Attribute)
+    assert prop.kind == 'Property'
+    assert prop.docstring == """For sale."""
+    assert type2str(prop.annotation) == 'str'
+
+    oldschool = C.contents['oldschool']
+    assert isinstance(oldschool, model.Attribute)
+    assert oldschool.kind == 'Property'
+    assert isinstance(oldschool.parsed_docstring, ParsedEpytextDocstring)
+    assert unwrap(oldschool.parsed_docstring) == """For rent."""
+    assert flatten(format_summary(oldschool)) == '<span>For rent.</span>'
+    assert isinstance(oldschool.parsed_type, ParsedEpytextDocstring)
+    assert str(unwrap(oldschool.parsed_type)) == 'string'
+    fields = oldschool.parsed_docstring.fields
+    assert len(fields) == 1
+    assert fields[0].tag() == 'see'
+
+
+@systemcls_param
+def test_property_setter(systemcls: Type[model.System], capsys: CapSys) -> None:
+    """Property setter and deleter methods are renamed, so they don't replace
+    the property itself.
+    """
+    mod = fromText('''
+    class C:
+        @property
+        def prop(self):
+            """Getter."""
+        @prop.setter
+        def prop(self, value):
+            """Setter."""
+        @prop.deleter
+        def prop(self):
+            """Deleter."""
+    ''', modname='mod', systemcls=systemcls)
+    C = mod.contents['C']
+
+    getter = C.contents['prop']
+    assert isinstance(getter, model.Attribute)
+    assert getter.kind == 'Property'
+    assert getter.docstring == """Getter."""
+
+    setter = C.contents['prop.setter']
+    assert isinstance(setter, model.Function)
+    assert setter.kind == 'Method'
+    assert setter.docstring == """Setter."""
+
+    deleter = C.contents['prop.deleter']
+    assert isinstance(deleter, model.Function)
+    assert deleter.kind == 'Method'
+    assert deleter.docstring == """Deleter."""
+
+
+@systemcls_param
+def test_property_custom(systemcls: Type[model.System], capsys: CapSys) -> None:
+    """Any custom decorator with a name ending in 'property' makes a method
+    into a property getter.
+    """
+    mod = fromText('''
+    class C:
+        @deprecate.deprecatedProperty(incremental.Version("Twisted", 18, 7, 0))
+        def processes(self):
+            return {}
+        @async_property
+        async def remote_value(self):
+            return await get_remote_value()
+    ''', modname='mod', systemcls=systemcls)
+    C = mod.contents['C']
+
+    deprecated = C.contents['processes']
+    assert isinstance(deprecated, model.Attribute)
+    assert deprecated.kind == 'Property'
+
+    async_prop = C.contents['remote_value']
+    assert isinstance(async_prop, model.Attribute)
+    assert async_prop.kind == 'Property'
+
+
+@pytest.mark.parametrize('decoration', ('classmethod', 'staticmethod'))
+@systemcls_param
+def test_property_conflict(
+        decoration: str, systemcls: Type[model.System], capsys: CapSys
+        ) -> None:
+    """Warn when a function is decorated as both property and class/staticmethod.
+    These decoration combinations do not create class/static properties.
+    """
+    mod = fromText(f'''
+    class C:
+        @{decoration}
+        @property
+        def prop():
+            raise NotImplementedError
+    ''', modname='mod', systemcls=systemcls)
+    C = mod.contents['C']
+    assert C.contents['prop'].kind == 'Property'
+    captured = capsys.readouterr().out
+    assert captured == f"mod:3: mod.C.prop is both property and {decoration}\n"
+
+@systemcls_param
+def test_ignore_function_contents(systemcls: Type[model.System]) -> None:
     mod = fromText('''
     def outer():
         """Outer function."""
